from __future__ import annotations

import numpy as np
import xarray as xr

from xclim.ensembles import fractional_uncertainty, hawkins_sutton, lafferty_sriver
from xclim.ensembles._filters import _concat_hist, _model_in_all_scens, _single_member


def test_hawkins_sutton_smoke(open_dataset):
    """Just a smoke test."""
    dims = {"run": "member", "scen": "scenario"}
    da = (
        open_dataset("uncertainty_partitioning/cmip5_pr_global_mon.nc")
        .pr.sel(time=slice("1950", None))
        .rename(dims)
    )
    da1 = _model_in_all_scens(da)
    dac = _concat_hist(da1, scenario="historical")
    das = _single_member(dac)
    hawkins_sutton(das)


def test_hawkins_sutton_synthetic(random):
    """Test logic of Hawkins-Sutton's implementation using synthetic data."""
    # Time, scenario, model
    # Here the scenarios don't change over time, so there should be no model variability (since it's relative to the
    # reference period.
    sm = np.arange(10, 41, 10)  # Scenario mean
    mm = np.arange(-6, 7, 1)  # Model mean
    mean = mm[np.newaxis, :] + sm[:, np.newaxis]

    # Natural variability
    r = random.standard_normal((4, 13, 60))

    x = r + mean[:, :, np.newaxis]
    time = xr.date_range("1970-01-01", periods=60, freq="YE")
    da = xr.DataArray(x, dims=("scenario", "model", "time"), coords={"time": time})
    m, v = hawkins_sutton(da)
    # Mean uncertainty over time
    vm = v.mean(dim="time")

    # Check that the mean relative to the baseline is zero
    np.testing.assert_array_almost_equal(m.mean(dim="time"), 0, decimal=1)

    # Check that the scenario uncertainty is zero
    np.testing.assert_array_almost_equal(vm.sel(uncertainty="scenario"), 0, decimal=1)

    # Check that model uncertainty > variability
    assert vm.sel(uncertainty="model") > vm.sel(uncertainty="variability")

    # Smoke test with polynomial of order 2
    fit = da.polyfit(dim="time", deg=2, skipna=True)
    sm = xr.polyval(coord=da.time, coeffs=fit.polyfit_coefficients).where(da.notnull())
    hawkins_sutton(da, sm=sm)

    # Test with a multiplicative variable and time evolving scenarios
    r = random.standard_normal((4, 13, 60)) + np.arange(60)
    x = r + mean[:, :, np.newaxis]
    da = xr.DataArray(x, dims=("scenario", "model", "time"), coords={"time": time})
    m, v = hawkins_sutton(da, kind="*")
    su = v.sel(uncertainty="scenario")
    # We expect the scenario uncertainty to grow over time
    # The scenarios all have the same absolute slope, but since their reference mean is different, the relative increase
    # is not the same and this creates a spread over time across "relative" scenarios.
    assert (
        su.sel(time=slice("2020", None)).mean()
        > su.sel(time=slice("2000", "2010")).mean()
    )


def test_lafferty_sriver_synthetic(random):
    """Test logic of Lafferty & Sriver's implementation using synthetic data."""
    # Time, scenario, model, downscaling
    # Here the scenarios don't change over time, so there should be no model variability (since it's relative to the
    # reference period.
    sm = np.arange(10, 41, 10)  # Scenario mean (4)
    mm = np.arange(-6, 7, 1)  # Model mean (13)
    dm = np.arange(-2, 3, 1)  # Downscaling mean (5)
    mean = (
        dm[np.newaxis, np.newaxis, :]
        + mm[np.newaxis, :, np.newaxis]
        + sm[:, np.newaxis, np.newaxis]
    )

    # Natural variability
    r = random.standard_normal((4, 13, 5, 60))

    x = r + mean[:, :, :, np.newaxis]
    time = xr.date_range("1970-01-01", periods=60, freq="YE")
    da = xr.DataArray(
        x, dims=("scenario", "model", "downscaling", "time"), coords={"time": time}
    )
    m, v = lafferty_sriver(da)
    # Mean uncertainty over time
    vm = v.mean(dim="time")

    # Check that the mean uncertainty
    np.testing.assert_array_almost_equal(m.mean(dim="time"), 25, decimal=1)

    # Check that model uncertainty > variability
    assert vm.sel(uncertainty="model") > vm.sel(uncertainty="variability")

    # Smoke test with polynomial of order 2
    fit = da.polyfit(dim="time", deg=2, skipna=True)
    sm = xr.polyval(coord=da.time, coeffs=fit.polyfit_coefficients).where(da.notnull())
    lafferty_sriver(da, sm=sm)


def test_lafferty_sriver(lafferty_sriver_ds):
    _g, u = lafferty_sriver(lafferty_sriver_ds.tas)

    fu = fractional_uncertainty(u)

    # Assertions based on expected results from
    # https://github.com/david0811/lafferty-sriver_2023_npjCliAtm/blob/main/unit_test/unit_test_check.ipynb
    assert fu.sel(time="2020", uncertainty="downscaling") > fu.sel(
        time="2020", uncertainty="model"
    )
    assert fu.sel(time="2020", uncertainty="variability") > fu.sel(
        time="2020", uncertainty="scenario"
    )
    assert (
        fu.sel(time="2090", uncertainty="scenario").data
        > fu.sel(time="2020", uncertainty="scenario").data
    )
    assert (
        fu.sel(time="2090", uncertainty="downscaling").data
        < fu.sel(time="2020", uncertainty="downscaling").data
    )

    def graph():
        """Return graphic like in https://github.com/david0811/lafferty-sriver_2023_npjCliAtm/blob/main/unit_test/unit_test_check.ipynb"""
        from matplotlib import pyplot as plt

        udict = {
            "Scenario": fu.sel(uncertainty="scenario").to_numpy().flatten(),
            "Model": fu.sel(uncertainty="model").to_numpy().flatten(),
            "Downscaling": fu.sel(uncertainty="downscaling").to_numpy().flatten(),
            "Variability": fu.sel(uncertainty="variability").to_numpy().flatten(),
        }

        _fig, ax = plt.subplots()
        ax.stackplot(
            np.arange(2015, 2101),
            udict.values(),
            labels=udict.keys(),
            alpha=1,
            colors=["#00CC89", "#6869B3", "#CC883C", "#FFFF99"],
            edgecolor="white",
            lw=1.5,
        )
        ax.set_xlim([2020, 2095])
        ax.set_ylim([0, 100])
        ax.legend(loc="upper left")
        plt.show()

    # graph()



from __future__ import annotations

from xclim import indices as xci
from xclim.core.indicator import Daily, ResamplingIndicatorWithIndexing
from xclim.indicators.atmos._conversion import Converter  # noqa

__all__ = [
    "blowing_snow",
    "snd_days_above",
    "snd_max_doy",
    "snd_season_end",
    "snd_season_length",
    "snd_season_start",
    "snd_storm_days",
    "snd_to_snw",
    "snow_depth",
    "snow_melt_we_max",
    "snw_days_above",
    "snw_max",
    "snw_max_doy",
    "snw_season_end",
    "snw_season_length",
    "snw_season_start",
    "snw_storm_days",
    "snw_to_snd",
]


# We need to declare the base class here so the `land` module is detected automatically.
class Snow(Daily):
    """Indicators dealing with snow variables."""

    keywords = "snow"


class SnowWithIndexing(ResamplingIndicatorWithIndexing):
    """Indicators dealing with snow variables, allowing indexing."""

    src_freq = "D"
    keywords = "snow"


snd_season_length = SnowWithIndexing(
    identifier="snd_season_length",
    units="days",
    long_name="Snow cover duration",
    description=(
        "The duration of the snow season, starting with at least {window} days with snow depth above {thresh} "
        "and ending with at least {window} days with snow depth under {thresh}."
    ),
    compute=xci.snd_season_length,
)

snw_season_length = SnowWithIndexing(
    identifier="snw_season_length",
    units="days",
    long_name="Snow cover duration",
    description=(
        "The duration of the snow season, starting with at least {window} days with snow amount above {thresh} "
        "and ending with at least {window} days with snow amount under {thresh}."
    ),
    compute=xci.snw_season_length,
)

snd_season_start = Snow(
    identifier="snd_season_start",
    standard_name="day_of_year",
    long_name="Start date of continuous snow depth cover",
    description="Day of year when snow depth is above or equal to {thresh} for {window} consecutive days.",
    abstract="The first date on which snow depth is greater than or equal to a given threshold "
    "for a given number of consecutive days.",
    units="",
    compute=xci.snd_season_start,
)

snw_season_start = Snow(
    identifier="snw_season_start",
    standard_name="day_of_year",
    long_name="Start date of continuous snow amount cover",
    description="Day of year when snow amount is above or equal to {thresh} for {window} consecutive days.",
    abstract="The first date on which snow amount is greater than or equal to a given threshold "
    "for a given number of consecutive days.",
    units="",
    compute=xci.snw_season_start,
)

snd_season_end = Snow(
    identifier="snd_season_end",
    standard_name="day_of_year",
    long_name="End date of continuous snow depth cover",
    description="Day of year when snow depth is below {thresh} for {window} consecutive days.",
    abstract="The first date on which snow depth is below a given threshold for a given number of consecutive days.",
    units="",
    compute=xci.snd_season_end,
)

snw_season_end = Snow(
    identifier="snw_season_end",
    standard_name="day_of_year",
    long_name="End date of continuous snow amount cover",
    description="Day of year when snow amount is below {thresh} for {window} consecutive days.",
    abstract="The first date on which snow amount is below a given threshold for a given number of consecutive days.",
    units="",
    compute=xci.snw_season_end,
)

snd_max_doy = SnowWithIndexing(
    title="Day of year of maximum snow depth",
    identifier="snd_max_doy",
    standard_name="day_of_year",
    var_name="{freq}_snd_max_doy",
    long_name="Day of the year when snow depth reaches its maximum value",
    description="The {freq} day of the year when snow depth reaches its maximum value.",
    abstract="Day of the year when snow depth reaches its maximum value.",
    units="",
    _partial=True,
    compute=xci.snd_max_doy,
)

snow_melt_we_max = Snow(
    title="Maximum snow melt",
    identifier="snow_melt_we_max",
    standard_name="change_over_time_in_surface_snow_amount",
    var_name="{freq}_snow_melt_we_max",
    long_name="Maximum snow melt",
    description="The {freq} maximum negative change in melt amount over {window} days.",
    abstract="The water equivalent of the maximum snow melt.",
    units="kg m-2",
    compute=xci.snow_melt_we_max,
)

snw_max = SnowWithIndexing(
    title="Maximum snow amount",
    identifier="snw_max",
    standard_name="surface_snow_amount",
    var_name="{freq}_snw_max",
    long_name="Maximum snow amount equivalent",
    description="The {freq} maximum snow amount equivalent on the surface.",
    abstract="The maximum snow amount equivalent on the surface.",
    units="kg m-2",
    compute=xci.snw_max,
)

snw_max_doy = SnowWithIndexing(
    title="Day of year of maximum snow amount",
    identifier="snw_max_doy",
    standard_name="day_of_year",
    var_name="{freq}_snw_max_doy",
    long_name="Day of year of maximum daily snow amount equivalent",
    description="The {freq} day of year when snow amount equivalent on the surface reaches its maximum.",
    abstract="The day of year when snow amount equivalent on the surface reaches its maximum.",
    units="",
    compute=xci.snw_max_doy,
)

melt_and_precip_max = Snow(
    title="Water equivalent maximum from precipitation and snow melt",
    identifier="melt_and_precip_max",
    var_name="{freq}_melt_and_precip_max",
    long_name="Water equivalent maximum from precipitation and snow melt",
    description="The {freq} maximum precipitation flux and negative change in snow amount over {window} days.",
    abstract="Maximum water input from precipitation flux and snow melt over a given window of days.",
    units="kg m-2",
    compute=xci.melt_and_precip_max,
)


snd_storm_days = SnowWithIndexing(
    title="Winter storm days",
    identifier="snd_storm_days",
    var_name="{freq}_snd_storm_days",
    long_name="Days with snowfall depth at or above a given threshold",
    description="The {freq} number of days with snowfall depth accumulation above {thresh}.",
    units="days",
    compute=xci.snd_storm_days,
)

snw_storm_days = SnowWithIndexing(
    title="Winter storm days",
    identifier="snw_storm_days",
    var_name="{freq}_snw_storm_days",
    long_name="Days with snowfall amount at or above a given threshold",
    description="The {freq} number of days with snowfall amount accumulation above {thresh}.",
    units="days",
    compute=xci.snw_storm_days,
)


blowing_snow = Snow(
    title="Blowing snow days",
    identifier="blowing_snow",
    var_name="{freq}_blowing_snow",
    long_name="Days with snowfall and wind speed at or above given thresholds",
    description="The {freq} number of days with snowfall over last {window} days above {snd_thresh} and wind speed "
    "above {sfcWind_thresh}.",
    abstract="The number of days with snowfall, snow depth, and windspeed over given thresholds for a period of days.",
    units="days",
    compute=xci.blowing_snow,
)

snow_depth = SnowWithIndexing(
    title="Mean snow depth",
    identifier="snow_depth",
    units="cm",
    standard_name="surface_snow_thickness",
    long_name="Mean of daily snow depth",
    description="The {freq} mean of daily mean snow depth.",
    abstract="Mean of daily snow depth.",
    cell_methods="time: mean over days",
    compute=xci.snow_depth,
)

snd_to_snw = Converter(
    title="Surface snow amount",
    identifier="snd_to_snw",
    units="kg m-2",
    standard_name="surface_snow_amount",
    long_name="Approximation of daily snow amount from snow depth and density",
    description="The approximation of daily snow amount from snow depth and density.",
    var_name="snw",
    compute=xci.snd_to_snw,
)

snw_to_snd = Converter(
    title="Surface snow depth",
    identifier="snw_to_snd",
    units="m",
    standard_name="surface_snow_thickness",
    long_name="Approximation of daily snow depth from snow amount and density",
    description="The approximation of daily snow depth from snow amount and density.",
    var_name="snd",
    compute=xci.snw_to_snd,
)


snd_days_above = SnowWithIndexing(
    title="Days with snow (depth)",
    identifier="snd_days_above",
    units="days",
    long_name="Number of days with snow",
    description="The {freq} number of days with snow depth greater than or equal to {thresh}.",
    abstract="Number of days when the snow depth is greater than or equal to a given threshold.",
    compute=xci.snd_days_above,
)

snw_days_above = SnowWithIndexing(
    title="Days with snow (amount)",
    identifier="snw_days_above",
    units="days",
    long_name="Number of days with snow",
    description="The {freq} number of days with snow amount greater than or equal to {thresh}.",
    abstract="Number of days when the snow amount is greater than or equal to a given threshold.",
    compute=xci.snw_days_above,
)



#!/usr/bin/env python
# Tests for `xclim` package.
#
# We want to test multiple things here:
#  - that data results are correct
#  - that metadata is correct and complete
#  - that missing data are handled appropriately
#  - that various calendar formats and supported
#  - that non-valid input frequencies or holes in the time series are detected
#
#
# For correctness, I think it would be useful to use a small dataset and run the original ICCLIM indicators on it,
# saving the results in a reference netcdf dataset. We could then compare the hailstorm output to this reference as
# a first line of defense.
from __future__ import annotations

import os
import sys
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from scipy.stats.mstats import mquantiles

from xclim import ensembles
from xclim.indices.stats import get_dist


# sklearn's KMeans doesn't accept the standard numpy Generator, so we create a special fixture for these tests
# This object is legacy and this fixture should only be used with KMeans, until they update their code to accept Generators instead.
# https://numpy.org/doc/stable/reference/random/legacy.html#numpy.random.RandomState
@pytest.fixture
def random_state():
    return np.random.RandomState(seed=list(map(ord, "𝕽𝔞𝖓𝔡𝖔𝔪")))


class TestEnsembleStats:
    def test_create_ensemble(self, open_dataset, ensemble_dataset_objects, nimbus):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files_simple"]:
            ds = open_dataset(n, decode_times=False)
            ds["time"] = xr.decode_cf(ds).time
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all)

        assert len(ens.realization) == len(ensemble_dataset_objects["nc_files_simple"])
        assert len(ens.time) == 151
        for i in np.arange(0, len(ens.realization)):
            np.testing.assert_array_equal(
                ens.isel(realization=i).tg_mean.values, ds_all[i].tg_mean.values
            )

        reals = [
            "_".join(Path(f).name.split("_")[1:4:2])
            for f in ensemble_dataset_objects["nc_files_simple"]
        ]
        ens1 = ensembles.create_ensemble(ds_all, realizations=reals)

        # Kinda a hack? Alternative is to open and rewrite in a temp folder.
        files = [nimbus.fetch(f) for f in ensemble_dataset_objects["nc_files_simple"]]
        ens2 = ensembles.create_ensemble(dict(zip(reals, files, strict=False)))
        xr.testing.assert_identical(ens1, ens2)

    def test_no_time(self, tmp_path, ensemble_dataset_objects, open_dataset):
        # create again using xr.Dataset objects
        f1 = Path(tmp_path / "notime")
        f1.mkdir()
        ds_all = []
        for n in ensemble_dataset_objects["nc_files"]:
            ds = open_dataset(n, decode_times=False)
            ds["time"] = xr.decode_cf(ds).time
            ds_all.append(ds.groupby(ds.time.dt.month).mean("time", keep_attrs=True))
            ds.groupby(ds.time.dt.month).mean("time", keep_attrs=True).to_netcdf(
                f1.joinpath(Path(n).name), engine="h5netcdf"
            )
        ens = ensembles.create_ensemble(ds_all)

        assert len(ens.realization) == len(ensemble_dataset_objects["nc_files"])

        in_ncs = list(Path(f1).glob("*.nc"))
        ens = ensembles.create_ensemble(in_ncs)
        assert len(ens.realization) == len(ensemble_dataset_objects["nc_files"])

    def test_create_unequal_times(self, ensemble_dataset_objects, open_dataset):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files"]:
            ds = open_dataset(n)
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all)

        assert len(ens.realization) == len(ds_all)
        assert ens.time.dt.year.min() == 1950
        assert ens.time.dt.year.max() == 2100
        assert len(ens.time) == 151

        ii = [
            i
            for i, s in enumerate(ensemble_dataset_objects["nc_files"])
            if "1970-2050" in s
        ]
        # assert padded with nans
        assert np.all(
            np.isnan(ens.tg_mean.isel(realization=ii).sel(time=ens.time.dt.year < 1970))
        )
        assert np.all(
            np.isnan(ens.tg_mean.isel(realization=ii).sel(time=ens.time.dt.year > 2050))
        )

        ens_mean = ens.tg_mean.mean(dim=["realization", "lon", "lat"], skipna=False)
        assert (
            ens_mean.where(~(np.isnan(ens_mean)), drop=True).time.dt.year.min() == 1970
        )
        assert (
            ens_mean.where(~(np.isnan(ens_mean)), drop=True).time.dt.year.max() == 2050
        )

    @pytest.mark.parametrize(
        "timegen,calkw",
        [(xr.cftime_range, {"calendar": "360_day"}), (pd.date_range, {})],
    )
    def test_create_unaligned_times(self, timegen, calkw):
        t1 = timegen("2000-01-01", periods=24, freq="ME", **calkw)
        t2 = timegen("2000-01-01", periods=24, freq="MS", **calkw)

        d1 = xr.DataArray(
            np.arange(24), dims=("time",), coords={"time": t1}, name="tas"
        )
        d2 = xr.DataArray(
            np.arange(24), dims=("time",), coords={"time": t2}, name="tas"
        )

        if t1.dtype != "O":
            ens = ensembles.create_ensemble((d1, d2))
            assert ens.time.size == 48
            np.testing.assert_equal(ens.isel(time=0), [np.nan, 0])
        ens = ensembles.create_ensemble((d1, d2), resample_freq="MS")

        assert ens.time.size == 24
        np.testing.assert_equal(ens.isel(time=0), [0, 0])

    @pytest.mark.parametrize("transpose", [False, True])
    def test_calc_perc(self, transpose, ensemble_dataset_objects, open_dataset):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files_simple"]:
            ds = open_dataset(n)
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all)

        if transpose:
            ens = ens.transpose()

        out1 = ensembles.ensemble_percentiles(ens, split=True)
        np.testing.assert_array_almost_equal(
            mquantiles(
                ens["tg_mean"].isel(time=0, lon=5, lat=5), 0.1, alphap=1, betap=1
            ),
            out1["tg_mean_p10"].isel(time=0, lon=5, lat=5),
        )
        np.testing.assert_array_almost_equal(
            mquantiles(
                ens["tg_mean"].isel(time=0, lon=5, lat=5), alphap=1, betap=1, prob=0.50
            ),
            out1["tg_mean_p50"].isel(time=0, lon=5, lat=5),
        )
        np.testing.assert_array_almost_equal(
            mquantiles(
                ens["tg_mean"].isel(time=0, lon=5, lat=5), alphap=1, betap=1, prob=0.90
            ),
            out1["tg_mean_p90"].isel(time=0, lon=5, lat=5),
        )

        # Specify method
        np.testing.assert_array_almost_equal(
            mquantiles(
                ens["tg_mean"].isel(time=0, lon=5, lat=5),
                alphap=0.5,
                betap=0.5,
                prob=0.90,
            ),
            ensembles.ensemble_percentiles(
                ens.isel(time=0, lon=5, lat=5), values=[90], method="hazen"
            ).tg_mean_p90,
        )

        assert np.all(out1["tg_mean_p90"] > out1["tg_mean_p50"])
        assert np.all(out1["tg_mean_p50"] > out1["tg_mean_p10"])

        out2 = ensembles.ensemble_percentiles(ens, values=(25, 75))
        assert np.all(out2["tg_mean_p75"] > out2["tg_mean_p25"])
        assert "Computation of the percentiles on" in out1.attrs["history"]

        out3 = ensembles.ensemble_percentiles(ens, split=False)
        xr.testing.assert_equal(
            out1["tg_mean_p10"], out3.tg_mean.sel(percentiles=10, drop=True)
        )

        weights = xr.DataArray(
            [1, 0.1, 3.5, 5], coords={"realization": ens.realization}
        )
        out4 = ensembles.ensemble_percentiles(ens, weights=weights)
        np.testing.assert_array_almost_equal(
            ens["tg_mean"].isel(time=0, lon=5, lat=5).weighted(weights).quantile(0.5),
            out4["tg_mean_p50"].isel(time=0, lon=5, lat=5),
        )
        np.testing.assert_array_almost_equal(
            ens["tg_mean"].isel(time=0, lon=5, lat=5).weighted(weights).quantile(0.1),
            out4["tg_mean_p10"].isel(time=0, lon=5, lat=5),
        )
        np.testing.assert_array_almost_equal(
            ens["tg_mean"].isel(time=0, lon=5, lat=5).weighted(weights).quantile(0.9),
            out4["tg_mean_p90"].isel(time=0, lon=5, lat=5),
        )
        assert np.all(out4["tg_mean_p90"] > out4["tg_mean_p10"])

    @pytest.mark.parametrize("keep_chunk_size", [False, True, None])
    def test_calc_perc_dask(
        self, keep_chunk_size, ensemble_dataset_objects, open_dataset
    ):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files_simple"]:
            ds = open_dataset(n)
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all)

        out2 = ensembles.ensemble_percentiles(
            ens.chunk({"time": 2}), keep_chunk_size=keep_chunk_size, split=False
        )
        out1 = ensembles.ensemble_percentiles(ens.load(), split=False)
        np.testing.assert_array_equal(out1["tg_mean"], out2["tg_mean"])

    def test_calc_perc_nans(self, ensemble_dataset_objects, open_dataset):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files_simple"]:
            ds = open_dataset(n)
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all).load()

        ens.tg_mean[2, 0, 5, 5] = np.nan
        ens.tg_mean[2, 7, 5, 5] = np.nan
        out1 = ensembles.ensemble_percentiles(ens, split=True)
        masked_arr = np.ma.fix_invalid(ens["tg_mean"][:, 0, 5, 5])
        np.testing.assert_array_almost_equal(
            mquantiles(masked_arr, 0.10, alphap=1, betap=1),
            out1["tg_mean_p10"][0, 5, 5],
        )
        masked_arr = np.ma.fix_invalid(ens["tg_mean"][:, 7, 5, 5])
        np.testing.assert_array_almost_equal(
            mquantiles(masked_arr, 0.10, alphap=1, betap=1),
            out1["tg_mean_p10"][7, 5, 5],
        )
        assert np.all(out1["tg_mean_p90"] > out1["tg_mean_p50"])
        assert np.all(out1["tg_mean_p50"] > out1["tg_mean_p10"])

    def test_calc_mean_std_min_max(self, ensemble_dataset_objects, open_dataset):
        ds_all = []
        for n in ensemble_dataset_objects["nc_files_simple"]:
            ds = open_dataset(n)
            ds_all.append(ds)
        ens = ensembles.create_ensemble(ds_all, multifile=False)

        out1 = ensembles.ensemble_mean_std_max_min(ens)
        np.testing.assert_array_equal(
            ens["tg_mean"][:, 0, 5, 5].mean(dim="realization"),
            out1.tg_mean_mean[0, 5, 5],
        )
        np.testing.assert_array_equal(
            ens["tg_mean"][:, 0, 5, 5].std(dim="realization"),
            out1.tg_mean_stdev[0, 5, 5],
        )
        np.testing.assert_array_equal(
            ens["tg_mean"][:, 0, 5, 5].max(dim="realization"), out1.tg_mean_max[0, 5, 5]
        )
        np.testing.assert_array_equal(
            ens["tg_mean"][:, 0, 5, 5].min(dim="realization"), out1.tg_mean_min[0, 5, 5]
        )
        assert "Computation of statistics on" in out1.attrs["history"]

        weights = xr.DataArray(
            [1, 0.1, 3.5, 5], coords={"realization": ens.realization}
        )
        out2 = ensembles.ensemble_mean_std_max_min(ens, weights=weights)
        values = ens["tg_mean"][:, 0, 5, 5]
        # Explicit float64 so numpy does the expected datatype promotion (change in numpy 2)
        np.testing.assert_array_equal(
            (
                values[0] * np.float64(1)
                + values[1] * np.float64(0.1)
                + values[2] * np.float64(3.5)
                + values[3] * np.float64(5)
            )
            / np.sum(weights),
            out2.tg_mean_mean[0, 5, 5],
        )
        np.testing.assert_array_equal(
            ens["tg_mean"][:, 0, 5, 5].weighted(weights).std(dim="realization"),
            out2.tg_mean_stdev[0, 5, 5],
        )
        np.testing.assert_array_equal(
            out1.tg_mean_max[0, 5, 5], out2.tg_mean_max[0, 5, 5]
        )
        np.testing.assert_array_equal(
            out1.tg_mean_min[0, 5, 5], out2.tg_mean_min[0, 5, 5]
        )

    @pytest.mark.parametrize(
        "aggfunc", [ensembles.ensemble_percentiles, ensembles.ensemble_mean_std_max_min]
    )
    def test_stats_min_members(self, ensemble_dataset_objects, open_dataset, aggfunc):
        ds_all = [open_dataset(n) for n in ensemble_dataset_objects["nc_files_simple"]]
        ens = ensembles.create_ensemble(ds_all).isel(lat=0, lon=0)
        ens = ens.where(ens.realization > 0)
        ens = xr.where((ens.realization == 1) & (ens.time.dt.year == 1950), np.nan, ens)

        def first(ds):
            return ds[list(ds.data_vars.keys())[0]]

        # Default, no masking
        out = first(aggfunc(ens))
        assert not out.isnull().any()

        # A number
        out = first(aggfunc(ens, min_members=3))
        # Only 1950 is null
        np.testing.assert_array_equal(
            out.isnull(), [True] + [False] * (ens.time.size - 1)
        )

        # Special value
        out = first(aggfunc(ens, min_members=None))
        # All null
        assert out.isnull().all()


@pytest.mark.slow
class TestEnsembleReduction:
    nc_file = os.path.join("EnsembleReduce", "TestEnsReduceCriteria.nc")

    def test_kmeans_rsqcutoff(self, open_dataset, random_state):
        pytest.importorskip("sklearn", minversion="0.24.1")
        ds = open_dataset(self.nc_file)

        # use random state variable to ensure consistent clustering in tests:
        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.5},
            random_state=random_state,
            make_graph=False,
        )

        assert ids == [4, 7, 10, 23]
        assert len(ids) == 4

        # Test max cluster option
        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.5},
            random_state=random_state,
            make_graph=False,
            max_clusters=3,
        )
        assert ids == [4, 7, 23]
        assert len(ids) == 3

    def test_kmeans_rsqopt(self, open_dataset, random_state):
        pytest.importorskip("sklearn", minversion="0.24.1")
        ds = open_dataset(self.nc_file)
        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_optimize": None},
            random_state=random_state,
            make_graph=False,
        )
        assert ids == [3, 4, 5, 7, 10, 11, 12, 13]

    def test_kmeans_nclust(self, open_dataset, random_state):
        ds = open_dataset(self.nc_file)

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"n_clusters": 4},
            random_state=random_state,
            make_graph=False,
        )
        assert ids == [4, 7, 10, 23]

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            ds.data,
            method={"n_clusters": 9},
            random_state=random_state,
            make_graph=False,
        )
        assert ids == [0, 3, 4, 6, 7, 10, 11, 12, 13]

    def test_kmeans_sampleweights(self, open_dataset, random_state):
        ds = open_dataset(self.nc_file)
        # Test sample weights
        sample_weights = np.ones(ds.data.shape[0])
        # boost weights for some sims
        sample_weights[[0, 20]] = 15

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.5},
            random_state=random_state,
            make_graph=False,
            sample_weights=sample_weights,
        )
        assert ids == [0, 20, 23]

        # RSQ optimize
        sample_weights = np.ones(ds.data.shape[0])
        # try zero weights
        sample_weights[[6, 18, 22]] = 0

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_optimize": None},
            random_state=random_state,
            make_graph=False,
            sample_weights=sample_weights,
        )

        assert ids == [4, 5, 7, 10, 11, 12, 13]

    def test_kmeans_variweights(self, open_dataset, random_state):
        pytest.importorskip("sklearn", minversion="0.24.1")
        ds = open_dataset(self.nc_file)
        # Test sample weights
        var_weights = np.ones(ds.data.shape[1])
        # reduce weights for some variables
        var_weights[3:] = 0.25

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.9},
            random_state=random_state,
            make_graph=False,
            variable_weights=var_weights,
        )
        assert ids == [1, 3, 8, 10, 13, 14, 16, 19, 20]

        # using RSQ optimize and try zero weights
        var_weights = np.ones(ds.data.shape[1])
        var_weights[[1, 4]] = 0

        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_optimize": None},
            random_state=random_state,
            make_graph=False,
            variable_weights=var_weights,
        )
        # Results here may change according to sklearn version, hence the *isin* instead of ==
        assert all(np.isin([12, 13, 16], ids))
        assert len(ids) == 6

    def test_kmeans_modelweights(self, open_dataset, random_state):
        ds = open_dataset(self.nc_file)
        # Test sample weights
        model_weights = np.ones(ds.data.shape[0])
        model_weights[[4, 7, 10, 23]] = 0

        # set to zero for some models that are selected in n_cluster test - these models should not be selected now
        [ids, cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"n_clusters": 4},
            random_state=random_state,
            make_graph=False,
            model_weights=model_weights,
        )

        for i in np.where(model_weights == 0)[0]:
            # as long as the cluster has more than one member the models w/ weight==0 should not be present
            if np.sum(cluster == cluster[i]) > 1:
                assert i not in ids

    @pytest.mark.skipif(
        "matplotlib.pyplot" not in sys.modules, reason="matplotlib.pyplot is required"
    )
    def test_kmeans_rsqcutoff_with_graphs(self, open_dataset, random_state):
        pytest.importorskip("sklearn", minversion="0.24.1")
        ds = open_dataset(self.nc_file)

        # use random state variable to ensure consistent clustering in tests:
        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.5},
            random_state=random_state,
            make_graph=True,
        )

        assert ids == [4, 7, 10, 23]

        # Test max cluster option
        [ids, _cluster, _fig_data] = ensembles.kmeans_reduce_ensemble(
            data=ds.data,
            method={"rsq_cutoff": 0.5},
            random_state=random_state,
            make_graph=True,
            max_clusters=3,
        )
        assert ids == [4, 7, 23]

    @pytest.mark.parametrize(
        "crit,num_select,expected",
        [
            ([0, 1], 5, [16, 19, 20, 15, 9]),
            ([0, 1], 4, [16, 19, 20, 15]),
            (np.arange(6), 8, [23, 19, 10, 14, 11, 8, 20, 3]),
            ([4, 5], 4, [15, 10, 14, 1]),
            ([4, 5], 1, [15]),
        ],
    )
    def test_kkz_simple(self, open_dataset, crit, num_select, expected):
        ens = open_dataset(self.nc_file)
        data = ens.data.isel(criteria=crit)

        selected = ensembles.kkz_reduce_ensemble(data, num_select)
        assert selected == expected

    def test_kkz_standardize(self, open_dataset):
        ens = open_dataset(self.nc_file)
        data = ens.data.isel(criteria=[1, 3, 5])

        sel_std = ensembles.kkz_reduce_ensemble(data, 4, standardize=True)
        sel_no = ensembles.kkz_reduce_ensemble(data, 4, standardize=False)
        assert sel_std == [23, 10, 19, 14]
        assert sel_no == [23, 1, 14, 10]

    def test_kkz_change_metric(self, open_dataset):
        # This test uses stupid values but is meant to test is kwargs are passed and if dist_method is used.
        ens = open_dataset(self.nc_file)
        data = ens.data.isel(criteria=[1, 3, 5])

        sel_euc = ensembles.kkz_reduce_ensemble(data, 4, dist_method="euclidean")
        sel_mah = ensembles.kkz_reduce_ensemble(
            data, 4, dist_method="mahalanobis", VI=np.arange(24)
        )
        assert sel_euc == [23, 10, 19, 14]
        assert sel_mah == [5, 3, 4, 0]

    def test_standardize_seuclidean(self, open_dataset):
        # This test the odd choice of standardizing data for a standardized distance metric
        ens = open_dataset(self.nc_file)
        data = ens.data
        for n in np.arange(1, len(data)):
            sel1 = ensembles.kkz_reduce_ensemble(
                data, n, dist_method="seuclidean", standardize=True
            )
            sel2 = ensembles.kkz_reduce_ensemble(
                data, n, dist_method="seuclidean", standardize=False
            )
            sel3 = ensembles.kkz_reduce_ensemble(
                data, n, dist_method="euclidean", standardize=True
            )
            assert sel1 == sel2
            assert sel1 == sel3

    def test_make_criteria(self, tas_series):
        ds = xr.Dataset(
            data_vars={
                "var_a": tas_series([0, 1, 2, 3]),
                "var_b": tas_series([0, 1, 2, 3]).expand_dims(lat=[45, 47]),
                "var_c": tas_series([0, 1, 2, 3]),
            }
        ).expand_dims(realization=["A", "B", "C"])

        crit = ensembles.make_criteria(ds)
        assert crit.dims == ("realization", "criteria")
        assert crit.criteria.size == 16
        uncrit = crit.unstack("criteria").to_dataset("variables")
        assert set(uncrit.data_vars.keys()) == {"var_a", "var_b", "var_c"}
        assert set(uncrit.var_a.dims) == {"realization", "lat", "time"}

        crit = ensembles.make_criteria(ds.var_b)
        assert crit.dims == ("realization", "criteria")
        assert crit.criteria.size == 8
        uncrit = crit.unstack("criteria")
        assert set(uncrit.dims) == {"realization", "lat", "time"}

        crit = ensembles.make_criteria(ds.where(ds.var_a > 0))
        assert crit.dims == ("realization", "criteria")
        assert crit.criteria.size == 12
        np.testing.assert_array_equal(crit.isnull().sum(), 0)
        np.testing.assert_array_equal(crit.min(), 1)
        uncrit = crit.unstack("criteria").to_dataset("variables")
        assert set(uncrit.dims) == {"realization", "lat", "time"}
        assert uncrit.time.size == 3


# ## Tests for Robustness ##
@pytest.fixture
def robust_data(random):
    norm = get_dist("norm")
    ref = np.tile(
        np.array(
            [
                norm.rvs(loc=274, scale=0.8, size=(40,), random_state=random)
                for r in range(4)
            ]
        ),
        (4, 1, 1),
    )
    fut = np.array(
        [
            [
                norm.rvs(loc=loc, scale=sc, size=(40,), random_state=random)
                for loc, sc in shps
            ]
            for shps in (
                [
                    (274.0, 0.7),
                    (274.0, 0.6),
                    (274.0, 0.7),
                    (275.6, 1.1),
                ],  # 3 no change, 1 positive change
                [
                    (272.5, 1.2),
                    (272.4, 0.8),
                    (275.5, 0.8),
                    (275.6, 1.1),
                ],  # 2 neg change
                [
                    (275.6, 0.8),
                    (275.8, 1.2),
                    (276.5, 0.8),
                    (277.6, 1.1),
                ],  # All pos change
                [
                    (np.nan, 0.3),
                    (np.nan, 1.2),
                    (275.5, 0.8),
                    (275.6, 1.1),
                ],  # Some NaN
            )
        ]
    )
    ref = xr.DataArray(ref, dims=("lon", "realization", "time"), name="tas")
    ref["time"] = xr.cftime_range("2000-01-01", periods=40, freq="YS")
    fut = xr.DataArray(fut, dims=("lon", "realization", "time"), name="tas")
    fut["time"] = xr.cftime_range("2040-01-01", periods=40, freq="YS")
    return ref, fut


@pytest.mark.parametrize(
    "test,exp_chng_frac,exp_pos_frac,exp_changed,kws",
    [
        (
            "ttest",
            [0.75, 1, 1, 1],
            [0.5, 0.5, 1, 1],
            [
                [False, True, True, True],
                [True, True, True, True],
                [True, True, True, True],
                [False, False, True, True],
            ],
            {},
        ),
        (
            "welch-ttest",
            [0.25, 1, 1, 1],
            [0.25, 0.5, 1, 1],
            [
                [False, False, False, True],
                [True, True, True, True],
                [True, True, True, True],
                [False, False, True, True],
            ],
            {},
        ),
        (
            "mannwhitney-utest",
            [0.5, 1, 1, 1],
            [0.25, 0.5, 1, 1],
            [
                [False, False, True, True],
                [True, True, True, True],
                [True, True, True, True],
                [False, False, True, True],
            ],
            {},
        ),
        (
            "brownforsythe-test",
            [0.25, 0.25, 0.25, 0],
            [0.25, 0.0, 0.25, 0],
            [
                [False, True, False, False],
                [True, False, False, False],
                [False, False, False, True],
                [False, False, False, False],
            ],
            {},
        ),
        (
            "ipcc-ar6-c",
            [0.25, 1, 1, 1],
            [0.25, 0.5, 1, 1],
            None,
            {},
        ),
        (
            "threshold",
            [0.25, 1, 1, 1],
            [0.25, 0.5, 1, 1],
            None,
            {"rel_thresh": 0.002},
        ),
        (
            "threshold",
            [0, 0, 0.5, 0],
            [0, 0, 0.5, 0],
            None,
            {"abs_thresh": 2},
        ),
        (
            None,
            [1, 1, 1, 1],
            [0.5, 0.5, 1, 1],
            [],
            {},
        ),
    ],
)
def test_robustness_fractions(
    robust_data, test, exp_chng_frac, exp_pos_frac, exp_changed, kws
):
    ref, fut = robust_data
    fracs = ensembles.robustness_fractions(fut, ref, test=test, **kws)

    assert fracs.changed.attrs["test"] == str(test)

    np.testing.assert_array_almost_equal(fracs.positive, [0.5, 0.5, 1, 1])
    np.testing.assert_array_almost_equal(fracs.agree, [0.5, 0.5, 1, 1])
    np.testing.assert_array_almost_equal(fracs.valid, [1, 1, 1, 0.5])
    np.testing.assert_array_almost_equal(fracs.changed, exp_chng_frac)
    np.testing.assert_array_almost_equal(fracs.changed_positive, exp_pos_frac)

    if "pvals" in fracs:
        # 0.05 is the default p_change parameter
        changed = fracs.pvals < 0.05
        np.testing.assert_array_almost_equal(changed, exp_changed)


def test_robustness_fractions_weighted(robust_data):
    ref, fut = robust_data
    weights = xr.DataArray([1, 0.1, 3.5, 5], coords={"realization": ref.realization})
    fracs = ensembles.robustness_fractions(fut, ref, test=None, weights=weights)
    assert fracs.changed.attrs["test"] == "None"

    np.testing.assert_array_equal(fracs.changed, [1, 1, 1, 1])
    np.testing.assert_array_almost_equal(
        fracs.changed_positive, [0.53125, 0.88541667, 1.0, 1.0]
    )


def test_robustness_fractions_delta(robust_data):
    delta = xr.DataArray([-2, 1, -2, -1, 0, 0], dims=("realization",))
    fracs = ensembles.robustness_fractions(delta, test="threshold", abs_thresh=1.5)
    np.testing.assert_array_equal(fracs.changed, [2 / 6])
    np.testing.assert_array_equal(fracs.changed_positive, [0.0])
    np.testing.assert_array_equal(fracs.positive, [1 / 6])
    np.testing.assert_array_equal(fracs.agree, [3 / 6])

    delta = xr.DataArray([-2, 1, -2, -1], dims=("realization",))
    weights = xr.DataArray([4, 3, 2, 1], dims=("realization",))
    fracs = ensembles.robustness_fractions(
        delta, test="threshold", abs_thresh=1.5, weights=weights
    )
    np.testing.assert_array_equal(fracs.changed, [0.6])
    np.testing.assert_array_equal(fracs.positive, [0.3])
    np.testing.assert_array_equal(fracs.agree, [0.7])


def test_robustness_fractions_empty():
    """Test that NaN is returned if input arrays are full of NaNs."""
    r = np.full((20, 10), np.nan)
    f = np.full((20, 10), np.nan)

    ref = xr.DataArray(r, dims=("realization", "time"), name="tas")
    fut = xr.DataArray(f, dims=("realization", "time"), name="tas")

    f = ensembles.robustness_fractions(fut, ref, test="ttest")
    assert np.all(np.isnan(f.changed))


def test_robustness_categories():
    lat = xr.DataArray([1, 2, 3, 4], dims=("lat",), attrs={"axis": "Y"}, name="lat")
    changed = xr.DataArray([0.5, 0.8, 1, 1], dims=("lat",), coords={"lat": lat})
    agree = xr.DataArray([1, 0.5, 0.5, 1], dims=("lat",), coords={"lat": lat})

    categories = ensembles.robustness_categories(changed, agree)
    np.testing.assert_array_equal(categories, [2, 3, 3, 1])
    assert categories.flag_values == [1, 2, 3]
    assert (
        categories.flag_meanings
        == "robust_signal no_change_or_no_signal conflicting_signal"
    )
    assert categories.lat.attrs["axis"] == "Y"


def test_robustness_coefficient():
    # High
    ref = xr.DataArray([274, 275, 274.5, 276, 274.3, 273.3], dims=("time",), name="tas")
    fut = xr.DataArray(
        [
            [277, 277.1, 278, 278.4, 278.1, 276.9],
            [275, 275.8, 276, 275.2, 276.2, 275.7],
        ],
        dims=("realization", "time"),
        name="tas",
    )
    R = ensembles.robustness_coefficient(fut, ref)
    np.testing.assert_almost_equal(R, 0.91972477)

    fut = xr.DataArray(
        [
            [277, 277.1, 278, 278.4, 278.1, 276.9],
            [274, 274.8, 273.7, 274.2, 273.9, 274.5],
        ],
        dims=("realization", "time"),
        name="tas",
    )
    R = ensembles.robustness_coefficient(fut, ref)
    np.testing.assert_almost_equal(R, 0.83743842)

    R = ensembles.robustness_coefficient(fut.to_dataset(), ref.to_dataset())
    np.testing.assert_almost_equal(R.tas, 0.83743842)



# pylint: disable=missing-kwoa
"""
Adjustment Methods
==================
"""
from __future__ import annotations

from importlib.util import find_spec
from inspect import signature
from typing import Any
from warnings import warn

import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

from xclim.core.calendar import get_calendar
from xclim.core.formatting import gen_call_string, update_history
from xclim.core.options import OPTIONS, SDBA_EXTRA_OUTPUT, set_options
from xclim.core.units import convert_units_to
from xclim.core.utils import uses_dask
from xclim.indices import stats

from ._adjustment import (
    dotc_adjust,
    dqm_adjust,
    dqm_train,
    eqm_train,
    extremes_adjust,
    extremes_train,
    loci_adjust,
    loci_train,
    mbcn_adjust,
    mbcn_train,
    npdf_transform,
    otc_adjust,
    qdm_adjust,
    qm_adjust,
    scaling_adjust,
    scaling_train,
)
from .base import Grouper, ParametrizableWithDataset, parse_group
from .processing import grouped_time_indexes
from .utils import (
    ADDITIVE,
    best_pc_orientation_full,
    best_pc_orientation_simple,
    equally_spaced_nodes,
    pc_matrix,
    rand_rot_matrix,
)

__all__ = [
    "LOCI",
    "OTC",
    "BaseAdjustment",
    "DetrendedQuantileMapping",
    "EmpiricalQuantileMapping",
    "ExtremeValues",
    "MBCn",
    "NpdfTransform",
    "PrincipalComponents",
    "QuantileDeltaMapping",
    "Scaling",
    "dOTC",
]


class BaseAdjustment(ParametrizableWithDataset):
    """Base class for adjustment objects.

    Children classes should implement the `train` and / or the `adjust` method.

    This base class defined the basic input and output checks. It should only be used for a real adjustment
    if neither `TrainAdjust` nor `Adjust` fit the algorithm.
    """

    _allow_diff_calendars = True
    _attribute = "_xclim_adjustment"

    def __init__(self, *args, _trained=False, **kwargs):
        if _trained:
            super().__init__(*args, **kwargs)
        else:
            raise ValueError(
                "As of xclim 0.29, Adjustment object should be initialized through their `train` or `adjust` methods."
            )

    @classmethod
    def _check_inputs(cls, *inputs, group):
        """Raise an error if there are chunks along the main dimension.

        Also raises if :py:attr:`BaseAdjustment._allow_diff_calendars` is False and calendars differ.
        """
        for inda in inputs:
            if uses_dask(inda) and len(inda.chunks[inda.get_axis_num(group.dim)]) > 1:
                raise ValueError(
                    f"Multiple chunks along the main adjustment dimension {group.dim} is not supported."
                )

        # All calendars used by the inputs
        calendars = {get_calendar(inda, group.dim) for inda in inputs}
        if not cls._allow_diff_calendars and len(calendars) > 1:
            raise ValueError(
                "Inputs are defined on different calendars,"
                f" this is not supported for {cls.__name__} adjustment."
            )

        # Check multivariate dimensions
        mvcrds = []
        for inda in inputs:
            for crd in inda.coords.values():
                if crd.attrs.get("is_variables", False):
                    mvcrds.append(crd)
        if mvcrds and (
            not all(mvcrds[0].equals(mv) for mv in mvcrds[1:])
            or len(mvcrds) != len(inputs)
        ):
            coords = {mv.name for mv in mvcrds}
            raise ValueError(
                f"Inputs have different multivariate coordinates: {', '.join(coords)}."
            )

        if group.prop == "dayofyear" and (
            "default" in calendars or "standard" in calendars
        ):
            warn(
                "Strange results could be returned when using `dayofyear` grouping "
                "on data defined in the 'proleptic_gregorian' calendar."
            )

    @classmethod
    def _harmonize_units(cls, *inputs, target: dict[str] | str | None = None):
        """Convert all inputs to the same units.

        If the target unit is not given, the units of the first input are used.

        Returns the converted inputs and the target units.
        """

        def _harmonize_units_multivariate(
            *_inputs, _dim, _target: dict[str] | None = None
        ):
            def __convert_units_to(_input_da, _internal_dim, _internal_target):
                varss = _input_da[_internal_dim].values
                input_units = {
                    v: _input_da[_internal_dim].attrs["_units"][iv]
                    for iv, v in enumerate(varss)
                }
                if input_units == _internal_target:
                    return _input_da
                input_standard_names = {
                    v: _input_da[_internal_dim].attrs["_standard_name"][iv]
                    for iv, v in enumerate(varss)
                }
                for iv, v in enumerate(varss):
                    _input_da.attrs["units"] = input_units[v]
                    _input_da.attrs["standard_name"] = input_standard_names[v]
                    _input_da[{_internal_dim: iv}] = convert_units_to(
                        _input_da[{_internal_dim: iv}],
                        _internal_target[v],
                        context="infer",
                    )
                    _input_da[_internal_dim].attrs["_units"][iv] = _internal_target[v]
                _input_da.attrs["units"] = ""
                _input_da.attrs.pop("standard_name")
                return _input_da

            if _target is None:
                if "_units" not in _inputs[0][_dim].attrs or any(
                    u is None for u in _inputs[0][_dim].attrs["_units"]
                ):
                    error_msg = (
                        "Units are missing in some or all of the stacked variables."
                        "The dataset stacked with `stack_variables` given as input should include units for every variable."
                    )
                    raise ValueError(error_msg)

                _target = {
                    v: _inputs[0][_dim].attrs["_units"][iv]
                    for iv, v in enumerate(_inputs[0][_dim].values)
                }

            # `__convert_units_to`` was changing the units of the 3rd dataset during the 2nd loop
            # This explicit loop is designed to avoid this
            _outputs = []
            original_units = list(
                [_inp[_dim].attrs["_units"].copy() for _inp in _inputs]
            )
            for _inp, units in zip(_inputs, original_units, strict=False):
                _inp[_dim].attrs["_units"] = units
                _outputs.append(
                    __convert_units_to(
                        _inp, _internal_dim=_dim, _internal_target=_target
                    )
                )
            return _outputs, _target

        for dim, crd in inputs[0].coords.items():
            if crd.attrs.get("is_variables"):
                return _harmonize_units_multivariate(*inputs, _dim=dim, _target=target)

        if target is None:
            target = inputs[0].units

        return (
            convert_units_to(inda, target, context="infer") for inda in inputs
        ), target

    @classmethod
    def _train(cls, ref, hist, **kwargs):
        raise NotImplementedError()

    def _adjust(self, sim, *args, **kwargs):
        raise NotImplementedError()


class TrainAdjust(BaseAdjustment):
    """Base class for adjustment objects obeying the train-adjust scheme.

    Children classes should implement these methods:

    - ``_train(ref, hist, **kwargs)``, classmethod receiving the training target and data,
      returning a training dataset and parameters to store in the object.

    - ``_adjust(sim, **kwargs)``, receiving the projected data and some arguments,
      returning the `scen` DataArray.

    """

    _allow_diff_calendars = True
    _attribute = "_xclim_adjustment"
    _repr_hide_params = ["hist_calendar", "train_units"]

    @classmethod
    def train(cls, ref: DataArray, hist: DataArray, **kwargs) -> TrainAdjust:
        r"""Train the adjustment object.

        Refer to the class documentation for the algorithm details.

        Parameters
        ----------
        ref : DataArray
            Training target, usually a reference time series drawn from observations.
        hist : DataArray
            Training data, usually a model output whose biases are to be adjusted.
        \*\*kwargs
            Algorithm-specific keyword arguments, see class doc.
        """
        kwargs = parse_group(cls._train, kwargs)
        skip_checks = kwargs.pop("skip_input_checks", False)

        if not skip_checks:
            if "group" in kwargs:
                cls._check_inputs(ref, hist, group=kwargs["group"])

            (ref, hist), train_units = cls._harmonize_units(ref, hist)
        else:
            train_units = ""

        ds, params = cls._train(ref, hist, **kwargs)
        obj = cls(
            _trained=True,
            hist_calendar=get_calendar(hist),
            train_units=train_units,
            **params,
        )
        obj.set_dataset(ds)
        return obj

    def adjust(self, sim: DataArray, *args, **kwargs):
        r"""Return bias-adjusted data.

        Refer to the class documentation for the algorithm details.

        Parameters
        ----------
        sim : DataArray
            Time series to be bias-adjusted, usually a model output.
        args : xr.DataArray
            Other DataArrays needed for the adjustment (usually none).
        \*\*kwargs
            Algorithm-specific keyword arguments, see class doc.
        """
        skip_checks = kwargs.pop("skip_input_checks", False)
        if not skip_checks:
            if "group" in self:
                self._check_inputs(sim, *args, group=self.group)

            (sim, *args), _ = self._harmonize_units(sim, *args, target=self.train_units)

        out = self._adjust(sim, *args, **kwargs)

        if isinstance(out, xr.DataArray):
            out = out.rename("scen").to_dataset()

        scen = out.scen

        # Keep attrs
        scen.attrs.update(sim.attrs)
        for name, crd in sim.coords.items():
            if name in scen.coords:
                scen[name].attrs.update(crd.attrs)
        params = gen_call_string("", **kwargs)[1:-1]  # indexing to remove added ( )
        infostr = f"{str(self)}.adjust(sim, {params})"
        scen.attrs["history"] = update_history(f"Bias-adjusted with {infostr}", sim)
        scen.attrs["bias_adjustment"] = infostr

        _is_multivariate = any(
            _crd.attrs.get("is_variables") for _crd in sim.coords.values()
        )
        if _is_multivariate is False:
            scen.attrs["units"] = self.train_units

        if OPTIONS[SDBA_EXTRA_OUTPUT]:
            return out
        return scen

    def set_dataset(self, ds: xr.Dataset):
        """Store an xarray dataset in the `ds` attribute.

        Useful with custom object initialization or if some external processing was performed.
        """
        super().set_dataset(ds)
        self.ds.attrs["adj_params"] = str(self)

    @classmethod
    def _train(cls, ref: DataArray, hist: DataArray, *kwargs):
        raise NotImplementedError()

    def _adjust(self, sim, **kwargs):
        raise NotImplementedError()


class Adjust(BaseAdjustment):
    """Adjustment with no intermediate trained object.

    Children classes should implement a `_adjust` classmethod taking as input the three DataArrays
    and returning the scen dataset/array.
    """

    @classmethod
    def adjust(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        sim: xr.DataArray | None = None,
        **kwargs,
    ) -> xr.Dataset:
        r"""Return bias-adjusted data. Refer to the class documentation for the algorithm details.

        Parameters
        ----------
        ref : DataArray
            Training target, usually a reference time series drawn from observations.
        hist : DataArray
            Training data, usually a model output whose biases are to be adjusted.
        sim : DataArray
            Time series to be bias-adjusted, usually a model output.
        \*\*kwargs
            Algorithm-specific keyword arguments, see class doc.

        Returns
        -------
        xr.Dataset
            The bias-adjusted Dataset.
        """
        if sim is None:
            sim = hist.copy()
            sim.attrs["_is_hist"] = True

        kwargs = parse_group(cls._adjust, kwargs)
        skip_checks = kwargs.pop("skip_input_checks", False)

        if not skip_checks:
            if "group" in kwargs:
                cls._check_inputs(ref, hist, sim, group=kwargs["group"])

            (ref, hist, sim), _ = cls._harmonize_units(ref, hist, sim)

        out: xr.Dataset | xr.DataArray = cls._adjust(ref, hist, sim=sim, **kwargs)

        if isinstance(out, xr.DataArray):
            out = out.rename("scen").to_dataset()

        scen = out.scen

        params = ", ".join([f"{k}={repr(v)}" for k, v in kwargs.items()])
        infostr = f"{cls.__name__}.adjust(ref, hist, sim, {params})"
        scen.attrs["history"] = update_history(f"Bias-adjusted with {infostr}", sim)
        scen.attrs["bias_adjustment"] = infostr

        _is_multivariate = any(
            [_crd.attrs.get("is_variables") for _crd in sim.coords.values()]
        )
        if _is_multivariate is False:
            scen.attrs["units"] = ref.units

        if OPTIONS[SDBA_EXTRA_OUTPUT]:
            return out
        return scen


class EmpiricalQuantileMapping(TrainAdjust):
    """Empirical Quantile Mapping bias-adjustment.

    Adjustment factors are computed between the quantiles of `ref` and `sim`.
    Values of `sim` are matched to the corresponding quantiles of `hist` and corrected accordingly.

    .. math::

      F^{-1}_{ref} (F_{hist}(sim))

    where :math:`F` is the cumulative distribution function (CDF) and `mod` stands for model data.

    Attributes
    ----------
    Train step

    nquantiles : int or 1d array of floats
        The number of quantiles to use. Two endpoints at 1e-6 and 1 - 1e-6 will be added.
        An array of quantiles [0, 1] can also be passed. Defaults to 20 quantiles.
    kind : {'+', '*'}
        The adjustment kind, either additive or multiplicative. Defaults to "+".
    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning an single adjustment group along dimension "time".
    adapt_freq_thresh : str | None
        Threshold for frequency adaptation. See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Default is None, meaning that frequency adaptation is not performed.

    Adjust step:

    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use when interpolating the adjustment factors. Defaults to "nearest".
    extrapolation : {'constant', 'nan'}
        The type of extrapolation to use. See :py:func:`xclim.sdba.utils.extrapolate_qm` for details. Defaults to "constant".

    References
    ----------
    :cite:cts:`sdba-deque_frequency_2007`
    """

    _allow_diff_calendars = False

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        nquantiles: int | np.ndarray = 20,
        kind: str = ADDITIVE,
        group: str | Grouper = "time",
        adapt_freq_thresh: str | None = None,
        jitter_under_thresh_value: str | None = None,
    ) -> tuple[xr.Dataset, dict[str, Any]]:
        if np.isscalar(nquantiles):
            quantiles = equally_spaced_nodes(nquantiles).astype(ref.dtype)
        else:
            quantiles = nquantiles.astype(ref.dtype)

        ds = eqm_train(
            xr.Dataset({"ref": ref, "hist": hist}),
            group=group,
            kind=kind,
            quantiles=quantiles,
            adapt_freq_thresh=adapt_freq_thresh,
            jitter_under_thresh_value=jitter_under_thresh_value,
        )

        ds.af.attrs.update(
            standard_name="Adjustment factors",
            long_name="Quantile mapping adjustment factors",
        )
        ds.hist_q.attrs.update(
            standard_name="Model quantiles",
            long_name="Quantiles of model on the reference period",
        )
        return ds, {"group": group, "kind": kind}

    def _adjust(self, sim, interp="nearest", extrapolation="constant"):
        return qm_adjust(
            xr.Dataset({"af": self.ds.af, "hist_q": self.ds.hist_q, "sim": sim}),
            group=self.group,
            interp=interp,
            extrapolation=extrapolation,
            kind=self.kind,
        ).scen


class DetrendedQuantileMapping(TrainAdjust):
    r"""Detrended Quantile Mapping bias-adjustment.

    The algorithm follows these steps, 1-3 being the 'train' and 4-6, the 'adjust' steps.

    1. A scaling factor that would make the mean of `hist` match the mean of `ref` is computed.
    2. `ref` and `hist` are normalized by removing the "dayofyear" mean.
    3. Adjustment factors are computed between the quantiles of the normalized `ref` and `hist`.
    4. `sim` is corrected by the scaling factor, and either normalized by "dayofyear" and  detrended group-wise
       or directly detrended per "dayofyear", using a linear fit (modifiable).
    5. Values of detrended `sim` are matched to the corresponding quantiles of normalized `hist` and corrected accordingly.
    6. The trend is put back on the result.

    .. math::

        F^{-1}_{ref}\left\{F_{hist}\left[\frac{\overline{hist}\cdot sim}{\overline{sim}}\right]\right\}\frac{\overline{sim}}{\overline{hist}}

    where :math:`F` is the cumulative distribution function (CDF) and :math:`\overline{xyz}` is the linear trend of the data.
    This equation is valid for multiplicative adjustment. Based on the DQM method of :cite:p:`sdba-cannon_bias_2015`.

    Parameters
    ----------
    Train step:

    nquantiles : int or 1d array of floats
        The number of quantiles to use. See :py:func:`~xclim.sdba.utils.equally_spaced_nodes`.
        An array of quantiles [0, 1] can also be passed. Defaults to 20 quantiles.
    kind : {'+', '*'}
        The adjustment kind, either additive or multiplicative. Defaults to "+".
    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning a single adjustment group along dimension "time".
    adapt_freq_thresh : str | None
        Threshold for frequency adaptation. See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Default is None, meaning that frequency adaptation is not performed.

    Adjust step:

    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use when interpolating the adjustment factors. Defaults to "nearest".
    detrend : int or BaseDetrend instance
        The method to use when detrending. If an int is passed, it is understood as a PolyDetrend (polynomial detrending) degree.
        Defaults to 1 (linear detrending)
    extrapolation : {'constant', 'nan'}
        The type of extrapolation to use. See :py:func:`xclim.sdba.utils.extrapolate_qm` for details. Defaults to "constant".

    References
    ----------
    :cite:cts:`sdba-cannon_bias_2015`

    """

    _allow_diff_calendars = False

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        nquantiles: int | np.ndarray = 20,
        kind: str = ADDITIVE,
        group: str | Grouper = "time",
        adapt_freq_thresh: str | None = None,
        jitter_under_thresh_value: str | None = None,
    ):
        if group.prop not in ["group", "dayofyear"]:
            warn(
                f"Using DQM with a grouping other than 'dayofyear' is not recommended (received {group.name})."
            )

        if np.isscalar(nquantiles):
            quantiles = equally_spaced_nodes(nquantiles).astype(ref.dtype)
        else:
            quantiles = nquantiles.astype(ref.dtype)

        ds = dqm_train(
            xr.Dataset({"ref": ref, "hist": hist}),
            group=group,
            quantiles=quantiles,
            kind=kind,
            adapt_freq_thresh=adapt_freq_thresh,
            jitter_under_thresh_value=jitter_under_thresh_value,
        )

        ds.af.attrs.update(
            standard_name="Adjustment factors",
            long_name="Quantile mapping adjustment factors",
        )
        ds.hist_q.attrs.update(
            standard_name="Model quantiles",
            long_name="Quantiles of model on the reference period",
        )
        ds.scaling.attrs.update(
            standard_name="Scaling factor",
            description="Scaling factor making the mean of hist match the one of hist.",
        )
        return ds, {"group": group, "kind": kind}

    def _adjust(
        self,
        sim,
        interp="nearest",
        extrapolation="constant",
        detrend=1,
    ):
        scen = dqm_adjust(
            self.ds.assign(sim=sim),
            interp=interp,
            extrapolation=extrapolation,
            detrend=detrend,
            group=self.group,
            kind=self.kind,
        ).scen
        # Detrending needs units.
        scen.attrs["units"] = sim.units
        return scen


class QuantileDeltaMapping(EmpiricalQuantileMapping):
    r"""Quantile Delta Mapping bias-adjustment.

    Adjustment factors are computed between the quantiles of `ref` and `hist`.
    Quantiles of `sim` are matched to the corresponding quantiles of `hist` and corrected accordingly.

    .. math::

        sim\frac{F^{-1}_{ref}\left[F_{sim}(sim)\right]}{F^{-1}_{hist}\left[F_{sim}(sim)\right]}

    where :math:`F` is the cumulative distribution function (CDF). This equation is valid for multiplicative adjustment.
    The algorithm is based on the "QDM" method of :cite:p:`sdba-cannon_bias_2015`.

    Parameters
    ----------
    Train step:

    nquantiles : int or 1d array of floats
        The number of quantiles to use. See :py:func:`~xclim.sdba.utils.equally_spaced_nodes`.
        An array of quantiles [0, 1] can also be passed. Defaults to 20 quantiles.
    kind : {'+', '*'}
        The adjustment kind, either additive or multiplicative. Defaults to "+".
    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning a single adjustment group along dimension "time".

    Adjust step:

    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use when interpolating the adjustment factors. Defaults to "nearest".
    extrapolation : {'constant', 'nan'}
        The type of extrapolation to use. See :py:func:`xclim.sdba.utils.extrapolate_qm` for details. Defaults to "constant".

    Extra diagnostics
    -----------------
    In adjustment:

    quantiles : The quantile of each value of `sim`. The adjustment factor is interpolated using this as the "quantile" axis on `ds.af`.

    References
    ----------
    :cite:cts:`sdba-cannon_bias_2015`
    """

    def _adjust(self, sim, interp="nearest", extrapolation="constant"):
        out = qdm_adjust(
            xr.Dataset({"sim": sim, "af": self.ds.af, "hist_q": self.ds.hist_q}),
            group=self.group,
            interp=interp,
            extrapolation=extrapolation,
            kind=self.kind,
        )
        if OPTIONS[SDBA_EXTRA_OUTPUT]:
            out.sim_q.attrs.update(long_name="Group-wise quantiles of `sim`.")
            return out
        return out.scen


class ExtremeValues(TrainAdjust):
    r"""Adjustment correction for extreme values.

    The tail of the distribution of adjusted data is corrected according to the bias between the parametric Generalized
    Pareto distributions of the simulated and reference data :cite:p:`sdba-roy_extremeprecip_2023`. The distributions are composed of the
    maximal values of clusters of "large" values.  With "large" values being those above `cluster_thresh`. Only extreme
    values, whose quantile within the pool of large values are above `q_thresh`, are re-adjusted. See `Notes`.

    This adjustment method should be considered experimental and used with care.

    Parameters
    ----------
    Train step :

    cluster_thresh : Quantity (str with units)
        The threshold value for defining clusters.
    q_thresh : float
        The quantile of "extreme" values, [0, 1[. Defaults to 0.95.
    ref_params :  xr.DataArray, optional
        Distribution parameters to use instead of fitting a GenPareto distribution on `ref`.

    Adjust step:

    scen : DataArray
        This is a second-order adjustment, so the adjust method needs the first-order
        adjusted timeseries in addition to the raw "sim".
    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use when interpolating the adjustment factors. Defaults to "linear".
    extrapolation : {'constant', 'nan'}
        The type of extrapolation to use. See :py:func:`~xclim.sdba.utils.extrapolate_qm` for details. Defaults to "constant".
    frac : float
        Fraction where the cutoff happens between the original scen and the corrected one.
        See Notes, ]0, 1]. Defaults to 0.25.
    power : float
        Shape of the correction strength, see Notes. Defaults to 1.0.

    Notes
    -----
    Extreme values are extracted from `ref`, `hist` and `sim` by finding all "clusters", i.e. runs of consecutive values
    above `cluster_thresh`. The `q_thresh`th percentile of these values is taken on `ref` and `hist` and becomes
    `thresh`, the extreme value threshold. The maximal value of each cluster, if it exceeds that new threshold, is taken
    and Generalized Pareto distributions are fitted to them, for both `ref` and `hist`. The probabilities associated
    with each of these extremes in `hist` is used to find the corresponding value according to `ref`'s distribution.
    Adjustment factors are computed as the bias between those new extremes and the original ones.

    In the adjust step, a Generalized Pareto distributions is fitted on the cluster-maximums of `sim` and it is used to
    associate a probability to each extreme, values over the `thresh` compute in the training, without the clustering.
    The adjustment factors are computed by interpolating the trained ones using these probabilities and the
    probabilities computed from `hist`.

    Finally, the adjusted values (:math:`C_i`) are mixed with the pre-adjusted ones (`scen`, :math:`D_i`) using the
    following transition function:

    .. math::

        V_i = C_i * \tau + D_i * (1 - \tau)

    Where :math:`\tau` is a function of sim's extreme values (unadjusted, :math:`S_i`)
    and of arguments ``frac`` (:math:`f`) and ``power`` (:math:`p`):

    .. math::

        \tau = \left(\frac{1}{f}\frac{S - min(S)}{max(S) - min(S)}\right)^p

    Code based on an internal Matlab source and partly ib the `biascorrect_extremes` function of the julia package
    "ClimateTools.jl" :cite:p:`sdba-roy_juliaclimateclimatetoolsjl_2021`.

    Because of limitations imposed by the lazy computing nature of the dask backend, it
    is not possible to know the number of cluster extremes in `ref` and `hist` at the
    moment the output data structure is created. This is why the code tries to estimate
    that number and usually overestimates it. In the training dataset, this translated
    into a `quantile` dimension that is too large and variables `af` and `px_hist` are
    assigned NaNs on extra elements. This has no incidence on the calculations
    themselves but requires more memory than is useful.

    References
    ----------
    :cite:cts:`sdba-roy_juliaclimateclimatetoolsjl_2021`
    :cite:cts:`sdba-roy_extremeprecip_2023`
    """

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        cluster_thresh: str,
        ref_params: xr.Dataset | None = None,
        q_thresh: float = 0.95,
    ):
        cluster_thresh = convert_units_to(cluster_thresh, ref, context="infer")

        # Approximation of how many "quantiles" values we will get:
        N = (1 - q_thresh) * ref.time.size * 1.05  # extra padding for safety

        # ref_params: cast nan to f32 not to interfere with map_blocks dtype parsing
        #   ref and hist are f32, we want to have f32 in the output.
        ds = extremes_train(
            xr.Dataset(
                {
                    "ref": ref,
                    "hist": hist,
                    "ref_params": ref_params or np.float32(np.nan),
                }
            ),
            q_thresh=q_thresh,
            cluster_thresh=cluster_thresh,
            dist=stats.get_dist("genpareto"),
            quantiles=np.arange(int(N)),
            group="time",
        )

        ds.px_hist.attrs.update(
            long_name="Probability of extremes in hist",
            description="Parametric probabilities of extremes in the common domain of hist and ref.",
        )
        ds.af.attrs.update(
            long_name="Extremes adjustment factor",
            description="Multiplicative adjustment factor of extremes from hist to ref.",
        )
        ds.thresh.attrs.update(
            long_name=f"{q_thresh * 100}th percentile extreme value threshold",
            description=f"Mean of the {q_thresh * 100}th percentile of large values (x > {cluster_thresh}) of ref and hist.",
        )

        return ds.drop_vars(["quantiles"]), {"cluster_thresh": cluster_thresh}

    def _adjust(
        self,
        sim: xr.DataArray,
        scen: xr.DataArray,
        *,
        frac: float = 0.25,
        power: float = 1.0,
        interp: str = "linear",
        extrapolation: str = "constant",
    ):
        # Quantiles coord : cheat and assign 0 - 1, so we can use `extrapolate_qm`.
        ds = self.ds.assign(
            quantiles=(np.arange(self.ds.quantiles.size) + 1)
            / (self.ds.quantiles.size + 1)
        )

        scen = extremes_adjust(
            ds.assign(sim=sim, scen=scen),
            cluster_thresh=self.cluster_thresh,
            dist=stats.get_dist("genpareto"),
            frac=frac,
            power=power,
            interp=interp,
            extrapolation=extrapolation,
            group="time",
        )

        return scen


class LOCI(TrainAdjust):
    r"""Local Intensity Scaling (LOCI) bias-adjustment.

    This bias adjustment method is designed to correct daily precipitation time series by considering wet and dry days
    separately :cite:p:`sdba-schmidli_downscaling_2006`.

    Multiplicative adjustment factors are computed such that the mean of `hist` matches the mean of `ref` for values
    above a threshold.

    The threshold on the training target `ref` is first mapped to `hist` by finding the quantile in `hist` having the same
    exceedance probability as thresh in `ref`. The adjustment factor is then given by

    .. math::

       s = \frac{\left \langle ref: ref \geq t_{ref} \right\rangle - t_{ref}}{\left \langle hist : hist \geq t_{hist} \right\rangle - t_{hist}}

    In the case of precipitations, the adjustment factor is the ratio of wet-days intensity.

    For an adjustment factor `s`, the bias-adjustment of `sim` is:

    .. math::

      sim(t) = \max\left(t_{ref} + s \cdot (hist(t) - t_{hist}), 0\right)

    Attributes
    ----------
    Train step:

    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning a single adjustment group along dimension "time".
    thresh : str
        The threshold in `ref` above which the values are scaled.

    Adjust step:

    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use then interpolating the adjustment factors. Defaults to "linear".

    References
    ----------
    :cite:cts:`sdba-schmidli_downscaling_2006`
    """

    _allow_diff_calendars = False

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        thresh: str,
        group: str | Grouper = "time",
    ):
        thresh = convert_units_to(thresh, ref)
        ds = loci_train(
            xr.Dataset({"ref": ref, "hist": hist}), group=group, thresh=thresh
        )
        ds.af.attrs.update(long_name="LOCI adjustment factors")
        ds.hist_thresh.attrs.update(long_name="Threshold over modeled data")
        return ds, {"group": group, "thresh": thresh}

    def _adjust(self, sim, interp="linear"):
        return loci_adjust(
            xr.Dataset(
                {"hist_thresh": self.ds.hist_thresh, "af": self.ds.af, "sim": sim}
            ),
            group=self.group,
            thresh=self.thresh,
            interp=interp,
        ).scen


class Scaling(TrainAdjust):
    """Scaling bias-adjustment.

    Simple bias-adjustment method scaling variables by an additive or multiplicative factor so that the mean of `hist`
    matches the mean of `ref`.

    Parameters
    ----------
    Train step:

    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning an single adjustment group along dimension "time".
    kind : {'+', '*'}
        The adjustment kind, either additive or multiplicative. Defaults to "+".

    Adjust step:

    interp : {'nearest', 'linear', 'cubic'}
        The interpolation method to use then interpolating the adjustment factors. Defaults to "nearest".
    """

    _allow_diff_calendars = False

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        group: str | Grouper = "time",
        kind: str = ADDITIVE,
    ):
        ds = scaling_train(
            xr.Dataset({"ref": ref, "hist": hist}), group=group, kind=kind
        )
        ds.af.attrs.update(long_name="Scaling adjustment factors")
        return ds, {"group": group, "kind": kind}

    def _adjust(self, sim, interp="nearest"):
        return scaling_adjust(
            xr.Dataset({"sim": sim, "af": self.ds.af}),
            group=self.group,
            interp=interp,
            kind=self.kind,
        ).scen


class PrincipalComponents(TrainAdjust):
    r"""Principal component adjustment.

    This bias-correction method maps model simulation values to the observation space through principal components
    :cite:p:`sdba-hnilica_multisite_2017`. Values in the simulation space (multiple variables, or multiple sites) can be
    thought of as coordinate along axes, such as variable, temperature, etc. Principal components (PC) are a
    linear combinations of the original variables where the coefficients are the eigenvectors of the covariance matrix.
    Values can then be expressed as coordinates along the PC axes. The method makes the assumption that bias-corrected
    values have the same coordinates along the PC axes of the observations. By converting from the observation PC space
    to the original space, we get bias corrected values. See `Notes` for a mathematical explanation.

    Warnings
    --------
    Be aware that *principal components* is meant here as the algebraic operation defining a coordinate system
    based on the eigenvectors, not statistical principal component analysis.

    Attributes
    ----------
    group : Union[str, Grouper]
        The main dimension and grouping information. See Notes.
        See :py:class:`xclim.sdba.base.Grouper` for details.
        The adjustment will be performed on each group independently.
        Default is "time", meaning a single adjustment group along dimension "time".
    best_orientation : {'simple', 'full'}
        Which method to use when searching for the best principal component orientation.
        See :py:func:`~xclim.sdba.utils.best_pc_orientation_simple` and
        :py:func:`~xclim.sdba.utils.best_pc_orientation_full`.
        "full" is more precise, but it is much slower.
    crd_dim : str
        The data dimension along which the multiple simulation space dimensions are taken.
        For a multivariate adjustment, this usually is "multivar", as returned by `sdba.stack_variables`.
        For a multisite adjustment, this should be the spatial dimension.
        The training algorithm currently doesn't support any chunking
        along either `crd_dim`. `group.dim` and `group.add_dims`.

    Notes
    -----
    The input data is understood as a set of N points in a :math:`M`-dimensional space.

    - :math:`M` is taken along `crd_dim`.

    - :math:`N` is taken along the dimensions given through `group` : (the main `dim` but also, if requested, the `add_dims` and `window`).

    The principal components (PC) of `hist` and `ref` are used to defined new coordinate systems, centered on their
    respective means. The training step creates a matrix defining the transformation from `hist` to `ref`:

    .. math::

      scen = e_{R} + \mathrm{\mathbf{T}}(sim - e_{H})

    Where:

    .. math::

      \mathrm{\mathbf{T}} = \mathrm{\mathbf{R}}\mathrm{\mathbf{H}}^{-1}

    :math:`\mathrm{\mathbf{R}}` is the matrix transforming from the PC coordinates computed on `ref` to the data
    coordinates. Similarly, :math:`\mathrm{\mathbf{H}}` is transform from the `hist` PC to the data coordinates
    (:math:`\mathrm{\mathbf{H}}` is the inverse transformation). :math:`e_R` and :math:`e_H` are the centroids of the
    `ref` and `hist` distributions respectively. Upon running the  `adjust` step, one may decide to use :math:`e_S`,
    the centroid of the `sim` distribution, instead of :math:`e_H`.

    References
    ----------
    :cite:cts:`sdba-hnilica_multisite_2017,sdba-alavoine_distinct_2022`

    """

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        crd_dim: str,
        best_orientation: str = "simple",
        group: str | Grouper = "time",
    ):
        all_dims = set(ref.dims + hist.dims)

        # Dimension name for the "points"
        lblP = xr.core.utils.get_temp_dimname(all_dims, "points")

        # Rename coord on ref, multiindex do not like conflicting coordinates names
        lblM = crd_dim
        lblR = xr.core.utils.get_temp_dimname(ref.dims, lblM + "_out")
        ref = ref.rename({lblM: lblR})

        # The real thing, acting on 2D numpy arrays
        def _compute_transform_matrix(reference, historical):
            """Return the transformation matrix converting simulation coordinates to observation coordinates."""
            # Get transformation matrix from PC coords to ref, dropping points with a NaN coord.
            ref_na = np.isnan(reference).any(axis=0)
            R = pc_matrix(reference[:, ~ref_na])
            # Get transformation matrix from PC coords to hist, dropping points with a NaN coord.
            hist_na = np.isnan(historical).any(axis=0)
            H = pc_matrix(historical[:, ~hist_na])
            # This step needs vectorize with dask, but vectorize doesn't work with dask, argh.
            # Invert to get transformation matrix from hist to PC coords.
            Hinv = np.linalg.inv(H)
            # Fancy tricks to choose the best orientation on each axis.
            # (using eigenvectors, the output axes orientation is undefined)
            if best_orientation == "simple":
                orient = best_pc_orientation_simple(R, Hinv)
            elif best_orientation == "full":
                orient = best_pc_orientation_full(
                    R, Hinv, reference.mean(axis=1), historical.mean(axis=1), historical
                )
            else:
                raise ValueError(
                    f"Unknown `best_orientation` method: {best_orientation}."
                )
            # Get transformation matrix
            return (R * orient) @ Hinv

        # The group wrapper
        def _compute_transform_matrices(ds, dim):
            """Apply `_compute_transform_matrix` along dimensions other than time and the variables to map."""
            # The multiple PC-space dimensions are along lblR and lblM
            # Matrix multiplication in xarray behaves as a dot product across
            # same-name dimensions, instead of reducing according to the dimension order,
            # as in numpy or normal maths.
            if len(dim) > 1:
                reference = ds.ref.stack({lblP: dim})
                historical = ds.hist.stack({lblP: dim})
            else:
                reference = ds.ref.rename({dim[0]: lblP})
                historical = ds.hist.rename({dim[0]: lblP})
            transformation = xr.apply_ufunc(
                _compute_transform_matrix,
                reference,
                historical,
                input_core_dims=[[lblR, lblP], [lblM, lblP]],
                output_core_dims=[[lblR, lblM]],
                vectorize=True,
                dask="parallelized",
                output_dtypes=[float],
            )
            return transformation

        # Transformation matrix, from model coords to ref coords.
        trans = group.apply(_compute_transform_matrices, {"ref": ref, "hist": hist})
        trans.attrs.update(long_name="Transformation from training to target spaces.")

        ref_mean = group.apply("mean", ref)  # Centroids of ref
        ref_mean.attrs.update(long_name="Centroid point of target.")

        hist_mean = group.apply("mean", hist)  # Centroids of hist
        hist_mean.attrs.update(long_name="Centroid point of training.")

        ds = xr.Dataset({"trans": trans, "ref_mean": ref_mean, "hist_mean": hist_mean})

        ds.attrs["_reference_coord"] = lblR
        ds.attrs["_model_coord"] = lblM
        return ds, {"group": group}

    def _adjust(self, sim):
        lblR = self.ds.attrs["_reference_coord"]
        lblM = self.ds.attrs["_model_coord"]

        vmean = self.group.apply("mean", sim)

        def _compute_adjust(ds, dim):
            """Apply the mapping transformation."""
            scenario = ds.ref_mean + ds.trans.dot((ds.sim - ds.vmean), [lblM])
            return scenario

        scen = (
            self.group.apply(
                _compute_adjust,
                {
                    "ref_mean": self.ds.ref_mean,
                    "trans": self.ds.trans,
                    "sim": sim,
                    "vmean": vmean,
                },
                main_only=True,
            )
            .rename({lblR: lblM})
            .rename("scen")
        )
        return scen


class NpdfTransform(Adjust):
    r"""N-dimensional probability density function transform.

    This adjustment object combines both training and adjust steps in the `adjust` class method.

    A multivariate bias-adjustment algorithm described by :cite:t:`sdba-cannon_multivariate_2018`, as part of the MBCn
    algorithm, based on a color-correction algorithm described by :cite:t:`sdba-pitie_n-dimensional_2005`.

    This algorithm in itself, when used with QuantileDeltaMapping, is NOT trend-preserving.
    The full MBCn algorithm includes a reordering step provided here by :py:func:`xclim.sdba.processing.reordering`.

    See notes for an explanation of the algorithm.

    Parameters
    ----------
    base : BaseAdjustment
        An univariate bias-adjustment class. This is untested for anything else than QuantileDeltaMapping.
    base_kws : dict, optional
        Arguments passed to the training of the univariate adjustment.
    n_escore : int
        The number of elements to send to the escore function. The default, 0, means all elements are included.
        Pass -1 to skip computing the escore completely.
        Small numbers result in less significant scores, but the execution time goes up quickly with large values.
    n_iter : int
        The number of iterations to perform. Defaults to 20.
    pts_dim : str
        The name of the "multivariate" dimension. Defaults to "multivar", which is the
        normal case when using :py:func:`xclim.sdba.base.stack_variables`.
    adj_kws : dict, optional
        Dictionary of arguments to pass to the adjust method of the univariate adjustment.
    rot_matrices : xr.DataArray, optional
        The rotation matrices as a 3D array ('iterations', <pts_dim>, <anything>), with shape (n_iter, <N>, <N>).
        If left empty, random rotation matrices will be automatically generated.

    Notes
    -----
    The historical reference (:math:`T`, for "target"), simulated historical (:math:`H`) and simulated projected (:math:`S`)
    datasets are constructed by stacking the timeseries of N variables together. The algorithm is broken into the
    following steps:

    1. Rotate the datasets in the N-dimensional variable space with :math:`\mathbf{R}`, a random rotation NxN matrix.

    .. math::

        \tilde{\mathbf{T}} = \mathbf{T}\mathbf{R} \
        \tilde{\mathbf{H}} = \mathbf{H}\mathbf{R} \
        \tilde{\mathbf{S}} = \mathbf{S}\mathbf{R}

    2. An univariate bias-adjustment :math:`\mathcal{F}` is used on the rotated datasets.
    The adjustments are made in additive mode, for each variable :math:`i`.

    .. math::

        \hat{\mathbf{H}}_i, \hat{\mathbf{S}}_i = \mathcal{F}\left(\tilde{\mathbf{T}}_i, \tilde{\mathbf{H}}_i, \tilde{\mathbf{S}}_i\right)

    3. The bias-adjusted datasets are rotated back.

    .. math::

        \mathbf{H}' = \hat{\mathbf{H}}\mathbf{R} \\
        \mathbf{S}' = \hat{\mathbf{S}}\mathbf{R}


    These three steps are repeated a certain number of times, prescribed by argument ``n_iter``. At each
    iteration, a new random rotation matrix is generated.

    The original algorithm :cite:p:`sdba-pitie_n-dimensional_2005`, stops the iteration when some distance score converges.
    Following cite:t:`sdba-cannon_multivariate_2018` and the MBCn implementation in :cite:t:`sdba-cannon_mbc_2020`, we
    instead fix the number of iterations.

    As done by cite:t:`sdba-cannon_multivariate_2018`, the distance score chosen is the "Energy distance" from
    :cite:t:`sdba-szekely_testing_2004`. (see: :py:func:`xclim.sdba.processing.escore`).

    The random matrices are generated following a method laid out by :cite:t:`sdba-mezzadri_how_2007`.

    This is only part of the full MBCn algorithm, see :ref:`notebooks/sdba:Statistical Downscaling and Bias-Adjustment`
    for an example on how to replicate the full method with xclim. This includes a standardization of the simulated data
    beforehand, an initial univariate adjustment and the reordering of those adjusted series according to the rank
    structure of the output of this algorithm.

    References
    ----------
    :cite:cts:`sdba-cannon_multivariate_2018,sdba-cannon_mbc_2020,sdba-pitie_n-dimensional_2005,sdba-mezzadri_how_2007,sdba-szekely_testing_2004`
    """

    @classmethod
    def _adjust(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        sim: xr.DataArray,
        *,
        base: TrainAdjust = QuantileDeltaMapping,
        base_kws: dict[str, Any] | None = None,
        n_escore: int = 0,
        n_iter: int = 20,
        pts_dim: str = "multivar",
        adj_kws: dict[str, Any] | None = None,
        rot_matrices: xr.DataArray | None = None,
    ) -> xr.Dataset:
        if base_kws is None:
            base_kws = {}
        if "kind" in base_kws:
            warn(
                f'The adjustment kind cannot be controlled when using {cls.__name__}, it defaults to "+".'
            )
        base_kws.setdefault("kind", "+")

        # Assuming sim has the same coords as hist
        # We get the safest new name of the rotated dim.
        rot_dim = xr.core.utils.get_temp_dimname(
            set(ref.dims).union(hist.dims).union(sim.dims), pts_dim + "_prime"
        )

        # Get the rotation matrices
        rot_matrices = rot_matrices or rand_rot_matrix(
            ref[pts_dim], num=n_iter, new_dim=rot_dim
        ).rename(matrices="iterations")

        # Call a map_blocks on the iterative function
        # Sadly, this is a bit too complicated for map_blocks, we'll do it by hand.
        escores_tmpl = xr.broadcast(
            ref.isel({pts_dim: 0, "time": 0}),
            hist.isel({pts_dim: 0, "time": 0}),
        )[0].expand_dims(iterations=rot_matrices.iterations)

        template = xr.Dataset(
            data_vars={
                "scenh": xr.full_like(hist, np.nan).rename(time="time_hist"),
                "scen": xr.full_like(sim, np.nan),
                "escores": escores_tmpl,
            }
        )

        # Input data, rename time dim on sim since it can't be aligned with ref or hist.
        ds = xr.Dataset(
            data_vars={
                "ref": ref.rename(time="time_hist"),
                "hist": hist.rename(time="time_hist"),
                "sim": sim,
                "rot_matrices": rot_matrices,
            }
        )

        kwargs = {
            "base": base,
            "base_kws": base_kws,
            "n_escore": n_escore,
            "n_iter": n_iter,
            "pts_dim": pts_dim,
            "adj_kws": adj_kws or {},
        }

        with set_options(sdba_extra_output=False):
            out = ds.map_blocks(npdf_transform, template=template, kwargs=kwargs)

        out = out.assign(rotation_matrices=rot_matrices)
        out.scenh.attrs["units"] = hist.units
        return out


class OTC(Adjust):
    r"""Optimal Transport Correction.

    Following :cite:t:`sdba-robin_2019`, this multivariate bias correction method finds the optimal transport
    mapping between simulated and observed data. The correction of every simulated data point is the observed
    point it is mapped to.

    See notes for an explanation of the algorithm.

    Parameters
    ----------
    bin_width : dict or float, optional
        Bin widths for specified dimensions if is dict.
        For all dimensions if float.
        Will be estimated with Freedman-Diaconis rule by default.
    bin_origin : dict or float, optional
        Bin origins for specified dimensions if is dict.
        For all dimensions if float.
        Default is 0.
    num_iter_max : int, optional
        Maximum number of iterations used in the earth mover distance algorithm.
        Default is 100_000_000.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    adapt_freq_thresh : dict or str, optional
        Threshold for frequency adaptation per variable.
        See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Frequency adaptation is not applied to missing variables if is dict.
        Applied to all variables if is string.
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated.
        Default is "max_distance".
        See notes for details.
    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning a single adjustment group along dimension "time".
    pts_dim : str
        The name of the "multivariate" dimension. Defaults to "multivar", which is the
        normal case when using :py:func:`xclim.sdba.base.stack_variables`.

    Notes
    -----
    The simulated and observed data sets :math:`X` and :math:`Y` are discretized and standardized using histograms
    whose bin length along dimension `v` is given by `bin_width[v]`. An optimal transport plan :math:`P^*` is found by solving
    the linear program

    .. math::

        \mathop{\arg\!\min}_{P} \langle P,C\rangle \\
        s.t. P\mathbf{1} = X \\
            P^T\mathbf{1} = Y \\
            P \geq 0

    where :math:`C_{ij}` is the squared euclidean distance between the bin at position :math:`i` of :math:`X`'s histogram and
    the bin at position :math:`j` of :math:`Y`'s.

    All data points belonging to input bin at position :math:`i` are then separately assigned to output bin at position :math:`j`
    with probability :math:`P_{ij}`. A transformation of bin positions can be applied before computing the distances :math:`C_{ij}`
    to make variables on different scales more evenly taken into consideration by the optimization step. Available transformations are

    - `normalization = 'standardize'` :
        .. math::

            i_v' = \frac{i_v - mean(i_v)}{std(i_v)} \quad\quad\quad j_v' = \frac{j_v - mean(j_v)}{std(j_v)}

    - `normalization = 'max_distance'` :
        .. math::

            i_v' = \frac{i_v}{max \{|i_v - j_v|\}} \quad\quad\quad j_v' = \frac{j_v}{max \{|i_v - j_v|\}}

        such that
            .. math::

                max \{|i_v' - j_v'|\} = max \{|i_w' - j_w'|\} = 1

    - `normalization = 'max_value'` :
        .. math::

            i_v' = \frac{i_v}{max\{i_v\}} \quad\quad\quad j_v' = \frac{j_v}{max\{j_v\}}

    for variables :math:`v, w`. Default is `'max_distance'`.

    Note that `POT <https://pythonot.github.io/>`__ must be installed to use this method.

    This implementation is strongly inspired by :cite:t:`sdba-robin_2021`.
    The differences from this implementation are :

    - `bin_width` and `bin_origin` are dictionaries or float
    - Freedman-Diaconis rule is used to find the bin width when unspecified, and fallbacks to Scott's rule when 0 is obtained
    - `jitter_inside_bins` argument
    - `adapt_freq_thresh` argument
    - `transform` argument
    - `group` argument
    - `pts_dim` argument

    References
    ----------
    :cite:cts:`sdba-robin_2019,sdba-robin_2021`
    """

    @classmethod
    def _adjust(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        bin_width: dict | float | None = None,
        bin_origin: dict | float | None = None,
        num_iter_max: int | None = 100_000_000,
        jitter_inside_bins: bool = True,
        adapt_freq_thresh: dict | str | None = None,
        normalization: str | None = "max_distance",
        group: str | Grouper = "time",
        pts_dim: str = "multivar",
        **kwargs,
    ) -> xr.DataArray:
        if find_spec("ot") is None:
            raise ImportError(
                "POT is required for OTC and dOTC. Please install with `pip install POT`."
            )

        if normalization not in [None, "standardize", "max_distance", "max_value"]:
            raise ValueError(
                "`transform` should be in [None, 'standardize', 'max_distance', 'max_value']."
            )

        sim = kwargs.pop("sim")
        if "_is_hist" not in sim.attrs:
            raise ValueError("OTC does not take a `sim` argument.")

        if isinstance(adapt_freq_thresh, str):
            adapt_freq_thresh = {v: adapt_freq_thresh for v in hist[pts_dim].values}
        if adapt_freq_thresh is not None:
            _, units = cls._harmonize_units(sim)
            for var, thresh in adapt_freq_thresh.items():
                adapt_freq_thresh[var] = str(
                    convert_units_to(thresh, units[var], context="hydro")
                )

        scen = otc_adjust(
            xr.Dataset({"ref": ref, "hist": hist}),
            bin_width=bin_width,
            bin_origin=bin_origin,
            num_iter_max=num_iter_max,
            jitter_inside_bins=jitter_inside_bins,
            adapt_freq_thresh=adapt_freq_thresh,
            normalization=normalization,
            group=group,
            pts_dim=pts_dim,
        ).scen

        if adapt_freq_thresh is not None:
            for var in adapt_freq_thresh.keys():
                adapt_freq_thresh[var] = adapt_freq_thresh[var] + " " + units[var]

        for d in scen.dims:
            if d != pts_dim:
                scen = scen.dropna(dim=d)

        return scen


class dOTC(Adjust):
    r"""dynamical Optimal Transport Correction.

    This method is the dynamical version of :py:class:`~xclim.sdba.adjustment.OTC`, as presented by :cite:t:`sdba-robin_2019`.
    The temporal evolution of the model is found for every point by mapping the historical to the future dataset with
    optimal transport. A mapping between historical and reference data is found in the same way, and the temporal evolution
    of model data is applied to their assigned reference.

    See notes for an explanation of the algorithm.

    This implementation is strongly inspired by :cite:t:`sdba-robin_2021`.

    Parameters
    ----------
    bin_width : dict or float, optional
        Bin widths for specified dimensions if is dict.
        For all dimensions if float.
        Will be estimated with Freedman-Diaconis rule by default.
    bin_origin : dict or float, optional
        Bin origins for specified dimensions if is dict.
        For all dimensions if float.
        Default is 0.
    num_iter_max : int, optional
        Maximum number of iterations used in the network simplex algorithm.
    cov_factor : {None, 'std', 'cholesky'}
        A rescaling of the temporal evolution before it is applied to the reference.
        Note that "cholesky" cannot be used if some variables are multiplicative.
        See notes for details.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    kind : dict or str, optional
        Keys are variable names and values are adjustment kinds, either additive or multiplicative.
        Unspecified dimensions are treated as "+".
        Applied to all variables if is string.
    adapt_freq_thresh : dict or str, optional
        Threshold for frequency adaptation per variable.
        See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Frequency adaptation is not applied to missing variables if is dict.
        Applied to all variables if is string.
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated
        in the optimal transport. Default is "max_distance".
        See :py:class:`~xclim.sdba.adjustment.OTC` for details.
    group : Union[str, Grouper]
        The grouping information. See :py:class:`xclim.sdba.base.Grouper` for details.
        Default is "time", meaning a single adjustment group along dimension "time".
    pts_dim : str
        The name of the "multivariate" dimension. Defaults to "multivar", which is the
        normal case when using :py:func:`xclim.sdba.base.stack_variables`.

    Notes
    -----
    The simulated historical, simulated future and observed data sets :math:`X0`, :math:`X1` and :math:`Y0` are
    discretized and standardized using histograms whose bin length along dimension `k` is given by `bin_width[k]`.
    Mappings between :math:`Y0` and :math:`X0` on the one hand and between :math:`X0` and :math:`X1` on the other
    are found by optimal transport (see :py:class:`~xclim.sdba.adjustment.OTC`). The latter mapping is used to
    compute the temporal evolution of model data. This evolution is computed additively or multiplicatively for
    each variable depending on its `kind`, and is applied to observed data with

    .. math::

        Y1_i & := Y0_i + D \cdot v_i \;\; or \\
        Y1_i & := Y0_i * D \cdot v_i

    where
        - :math:`v_i` is the temporal evolution of historical simulated point :math:`i \in X0` to :math:`j \in X1`
        - :math:`Y0_i` is the observed data mapped to :math:`i`
        - :math:`D` is a correction factor given by
            - :math:`I` if `cov_factor is None`
            - :math:`diag(\frac{\sigma_{Y0}}{\sigma_{X0}})` if `cov_factor = "std"`
            - :math:`\frac{Chol(Y0)}{Chol(X0)}` where :math:`Chol` is the Cholesky decomposition if `cov_factor = "cholesky"`
        - :math:`Y1_i` is the correction of the future simulated data mapped to :math:`i`.

    Note that `POT <https://pythonot.github.io/>`__ must be installed to use this method.

    This implementation is strongly inspired by :cite:t:`sdba-robin_2021`.
    The differences from this reference are :

    - `bin_width` and `bin_origin` are dictionaries or float.
    - Freedman-Diaconis rule is used to find the bin width when unspecified, and fallbacks to Scott's rule when 0 is obtained.
    - `jitter_inside_bins` argument
    - `adapt_freq_thresh` argument
    - `transform` argument
    - `group` argument
    - `pts_dim` argument
    - `kind` argument

    References
    ----------
    :cite:cts:`sdba-robin_2019,sdba-robin_2021`
    """

    @classmethod
    def _adjust(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        sim: xr.DataArray,
        *,
        bin_width: dict | float | None = None,
        bin_origin: dict | float | None = None,
        num_iter_max: int | None = 100_000_000,
        cov_factor: str | None = "std",
        jitter_inside_bins: bool = True,
        kind: dict | str | None = None,
        adapt_freq_thresh: dict | str | None = None,
        normalization: str | None = "max_distance",
        group: str | Grouper = "time",
        pts_dim: str = "multivar",
    ) -> xr.DataArray:
        if find_spec("ot") is None:
            raise ImportError(
                "POT is required for OTC and dOTC. Please install with `pip install POT`."
            )

        if isinstance(kind, str):
            kind = {v: kind for v in hist[pts_dim].values}
        if kind is not None and "*" in kind.values() and cov_factor == "cholesky":
            raise ValueError(
                "Multiplicative correction is not supported with `cov_factor` = 'cholesky'."
            )

        if cov_factor not in [None, "std", "cholesky"]:
            raise ValueError("`cov_factor` should be in [None, 'std', 'cholesky'].")

        if normalization not in [None, "standardize", "max_distance", "max_value"]:
            raise ValueError(
                "`normalization` should be in [None, 'standardize', 'max_distance', 'max_value']."
            )

        if isinstance(adapt_freq_thresh, str):
            adapt_freq_thresh = {v: adapt_freq_thresh for v in hist[pts_dim].values}
        if adapt_freq_thresh is not None:
            _, units = cls._harmonize_units(sim)
            for var, thresh in adapt_freq_thresh.items():
                adapt_freq_thresh[var] = str(
                    convert_units_to(thresh, units[var], context="hydro")
                )

        scen = dotc_adjust(
            xr.Dataset({"ref": ref, "hist": hist, "sim": sim}),
            bin_width=bin_width,
            bin_origin=bin_origin,
            num_iter_max=num_iter_max,
            cov_factor=cov_factor,
            jitter_inside_bins=jitter_inside_bins,
            kind=kind,
            adapt_freq_thresh=adapt_freq_thresh,
            normalization=normalization,
            group=group,
            pts_dim=pts_dim,
        ).scen

        if adapt_freq_thresh is not None:
            for var in adapt_freq_thresh.keys():
                adapt_freq_thresh[var] = adapt_freq_thresh[var] + " " + units[var]

        for d in scen.dims:
            if d != pts_dim:
                scen = scen.dropna(dim=d, how="all")

        return scen


class MBCn(TrainAdjust):
    r"""Multivariate bias correction function using the N-dimensional probability density function transform.

    A multivariate bias-adjustment algorithm described by :cite:t:`sdba-cannon_multivariate_2018`
    based on a color-correction algorithm described by :cite:t:`sdba-pitie_n-dimensional_2005`.

    This algorithm in itself, when used with QuantileDeltaMapping, is NOT trend-preserving.
    The full MBCn algorithm includes a reordering step provided here by :py:func:`xclim.sdba.processing.reordering`.

    See notes for an explanation of the algorithm.

    Attributes
    ----------
    Train step

    ref : xr.DataArray
        Reference dataset.
    hist : xr.DataArray
        Historical dataset.
    base_kws : dict, optional
        Arguments passed to the training in the npdf transform.
    adj_kws : dict, optional
        Arguments passed to the adjusting in the npdf transform.
    n_escore : int
        The number of elements to send to the escore function. The default, 0, means all elements are included.
        Pass -1 to skip computing the escore completely.
        Small numbers result in less significant scores, but the execution time goes up quickly with large values.
    n_iter : int
        The number of iterations to perform. Defaults to 20.
    pts_dim : str
        The name of the "multivariate" dimension. Defaults to "multivar", which is the
        normal case when using :py:func:`xclim.sdba.base.stack_variables`.
    rot_matrices: xr.DataArray, optional
        The rotation matrices as a 3D array ('iterations', <pts_dim>, <anything>), with shape (n_iter, <N>, <N>).
        If left empty, random rotation matrices will be automatically generated.

    Adjust step

    ref : xr.DataArray
        Target reference dataset also needed for univariate bias correction preceding npdf transform
    hist: xr.DataArray
        Source dataset also needed for univariate bias correction preceding npdf transform
    sim : xr.DataArray
        Source dataset to adjust.
    base : BaseAdjustment
        Bias-adjustment class used for the univariate bias correction.
    base_kws : dict, optional
        Arguments passed to the training in the univariate bias correction
    adj_kws : dict, optional
        Arguments passed to the adjusting in the univariate bias correction
    period_dim : str, optional
        Name of the period dimension used when stacking time periods of `sim`  using :py:func:`xclim.core.calendar.stack_periods`.
        If specified, the interpolation of the npdf transform is performed only once and applied on all periods simultaneously.
        This should be more performant, but also more memory intensive.

    Notes
    -----
    The historical reference (:math:`T`, for "target"), simulated historical (:math:`H`) and simulated projected (:math:`S`)
    datasets are constructed by stacking the timeseries of N variables together. The algorithm is broken into the
    following steps:

    Training (only npdf transform training)

    1. Standardize `ref` and `hist` (see ``xclim.sdba.processing.standardize``.)

    2. Rotate the datasets in the N-dimensional variable space with :math:`\mathbf{R}`, a random rotation NxN matrix.

    .. math::

        \tilde{\mathbf{T}} = \mathbf{T}\mathbf{R} \
        \tilde{\mathbf{H}} = \mathbf{H}\mathbf{R}

    3. QuantileDeltaMapping is used to perform bias adjustment  :math:`\mathcal{F}` on the rotated datasets.
    The adjustment factor is conserved for later use in the adjusting step. The adjustments are made in additive mode,
    for each variable :math:`i`.

    .. math::

        \hat{\mathbf{H}}_i, \hat{\mathbf{S}}_i = \mathcal{F}\left(\tilde{\mathbf{T}}_i, \tilde{\mathbf{H}}_i, \tilde{\mathbf{S}}_i\right)

    4. The bias-adjusted datasets are rotated back.

    .. math::

        \mathbf{H}' = \hat{\mathbf{H}}\mathbf{R} \\
        \mathbf{S}' = \hat{\mathbf{S}}\mathbf{R}

    5. Repeat steps 2,3,4  three steps ``n_iter`` times, i.e. the number of randomly generated rotation matrices.

    Adjusting

    1. Perform the same steps as in training, with `ref, hist` replaced with `sim`. Step 3. of the training is modified, here we
    simply reuse the adjustment factors previously found in the training step to bias correct the standardized `sim` directly.

    2. Using the original (unstandardized) `ref,hist, sim`, perform a univariate bias adjustment using the ``base_scen`` class
    on `sim`.

    3. Reorder the dataset found in step 2. according to the ranks of the dataset found in step 1.


    The original algorithm :cite:p:`sdba-pitie_n-dimensional_2005`, stops the iteration when some distance score converges.
    Following cite:t:`sdba-cannon_multivariate_2018` and the MBCn implementation in :cite:t:`sdba-cannon_mbc_2020`, we
    instead fix the number of iterations.

    As done by cite:t:`sdba-cannon_multivariate_2018`, the distance score chosen is the "Energy distance" from
    :cite:t:`sdba-szekely_testing_2004`. (see: :py:func:`xclim.sdba.processing.escore`).

    The random matrices are generated following a method laid out by :cite:t:`sdba-mezzadri_how_2007`.

    References
    ----------
    :cite:cts:`sdba-cannon_multivariate_2018,sdba-cannon_mbc_2020,sdba-pitie_n-dimensional_2005,sdba-mezzadri_how_2007,sdba-szekely_testing_2004`

    Notes
    -----
    Only  "time" and "time.dayofyear" (with a suitable window) are implemented as possible values for `group`.
    """

    @classmethod
    def _train(
        cls,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        base_kws: dict[str, Any] | None = None,
        adj_kws: dict[str, Any] | None = None,
        n_escore: int = -1,
        n_iter: int = 20,
        pts_dim: str = "multivar",
        rot_matrices: xr.DataArray | None = None,
    ):
        # set default values for non-specified parameters
        base_kws = base_kws if base_kws is not None else {}
        adj_kws = adj_kws if adj_kws is not None else {}
        base_kws.setdefault("nquantiles", 20)
        base_kws.setdefault("group", Grouper("time", 1))
        adj_kws.setdefault("interp", "nearest")
        adj_kws.setdefault("extrapolation", "constant")

        if np.isscalar(base_kws["nquantiles"]):
            base_kws["nquantiles"] = equally_spaced_nodes(base_kws["nquantiles"])
        if isinstance(base_kws["group"], str):
            base_kws["group"] = Grouper(base_kws["group"], 1)
        if base_kws["group"].name == "time.month":
            raise NotImplementedError(
                "Received `group==time.month` in `base_kws`. Monthly grouping is not currently supported in the MBCn class."
            )
        # stack variables and prepare rotations
        if rot_matrices is not None:
            if pts_dim != rot_matrices.attrs["crd_dim"]:
                raise ValueError(
                    f"`crd_dim` attribute of `rot_matrices` ({rot_matrices.attrs['crd_dim']}) does not correspond to `pts_dim` ({pts_dim})."
                )
        else:
            rot_dim = xr.core.utils.get_temp_dimname(
                set(ref.dims).union(hist.dims), pts_dim + "_prime"
            )
            rot_matrices = rand_rot_matrix(
                ref[pts_dim], num=n_iter, new_dim=rot_dim
            ).rename(matrices="iterations")
        pts_dims = [rot_matrices.attrs[d] for d in ["crd_dim", "new_dim"]]

        # time indices corresponding to group and windowed group
        # used to divide datasets as map_blocks or groupby would do
        _, gw_idxs = grouped_time_indexes(ref.time, base_kws["group"])

        # training, obtain adjustment factors of the npdf transform
        ds = xr.Dataset({"ref": ref, "hist": hist})
        params = {
            "quantiles": base_kws["nquantiles"],
            "interp": adj_kws["interp"],
            "extrapolation": adj_kws["extrapolation"],
            "pts_dims": pts_dims,
            "n_escore": n_escore,
        }
        out = mbcn_train(ds, rot_matrices=rot_matrices, gw_idxs=gw_idxs, **params)
        params["group"] = base_kws["group"]

        # postprocess
        out["rot_matrices"] = rot_matrices

        out.af_q.attrs.update(
            standard_name="Adjustment factors",
            long_name="Quantile mapping adjustment factors",
        )
        return out, params

    def _adjust(
        self,
        sim: xr.DataArray,
        ref: xr.DataArray,
        hist: xr.DataArray,
        *,
        base: TrainAdjust = QuantileDeltaMapping,
        base_kws_vars: dict[str, Any] | None = None,
        adj_kws: dict[str, Any] | None = None,
        period_dim=None,
    ):
        # set default values for non-specified parameters
        base_kws_vars = base_kws_vars or {}
        pts_dim = self.pts_dims[0]
        for v in sim[pts_dim].values:
            base_kws_vars.setdefault(v, {})
            base_kws_vars[v].setdefault("group", self.group)
            if isinstance(base_kws_vars[v]["group"], str):
                base_kws_vars[v]["group"] = Grouper(base_kws_vars[v]["group"], 1)
            if base_kws_vars[v]["group"] != self.group:
                raise ValueError(
                    f"`group` input in _train and _adjust must be the same."
                    f"Got {self.group} and {base_kws_vars[v]['group']}"
                )
            base_kws_vars[v].pop("group")

            base_kws_vars[v].setdefault("nquantiles", self.ds.af_q.quantiles.values)
            if np.isscalar(base_kws_vars[v]["nquantiles"]):
                base_kws_vars[v]["nquantiles"] = equally_spaced_nodes(
                    base_kws_vars[v]["nquantiles"]
                )
            if "is_variables" in sim[pts_dim].attrs:
                if self.train_units == "":
                    _, units = self._harmonize_units(sim)
                else:
                    units = self.train_units

                if "jitter_under_thresh_value" in base_kws_vars[v]:
                    base_kws_vars[v]["jitter_under_thresh_value"] = str(
                        convert_units_to(
                            base_kws_vars[v]["jitter_under_thresh_value"],
                            units[v],
                            context="hydro",
                        )
                    )
                if "adapt_freq_thresh" in base_kws_vars[v]:
                    base_kws_vars[v]["adapt_freq_thresh"] = str(
                        convert_units_to(
                            base_kws_vars[v]["adapt_freq_thresh"],
                            units[v],
                            context="hydro",
                        )
                    )

        adj_kws = adj_kws or {}
        adj_kws.setdefault("interp", self.interp)
        adj_kws.setdefault("extrapolation", self.extrapolation)

        g_idxs, gw_idxs = grouped_time_indexes(ref.time, self.group)
        ds = self.ds.copy()
        ds["g_idxs"] = g_idxs
        ds["gw_idxs"] = gw_idxs

        # adjust (adjust for npft transform, train/adjust for univariate bias correction)
        out = mbcn_adjust(
            ref=ref,
            hist=hist,
            sim=sim,
            ds=ds,
            pts_dims=self.pts_dims,
            interp=self.interp,
            extrapolation=self.extrapolation,
            base=base,
            base_kws_vars=base_kws_vars,
            adj_kws=adj_kws,
            period_dim=period_dim,
        )

        return out


try:
    import SBCK
except ImportError:  # noqa: S110
    # SBCK is not installed, we will not generate the SBCK classes.
    pass
else:

    class _SBCKAdjust(Adjust):
        sbck = None  # The method

        @classmethod
        def _adjust(cls, ref, hist, sim, *, multi_dim=None, **kwargs):
            # Check inputs
            fit_needs_sim = "X1" in signature(cls.sbck.fit).parameters
            for k, v in signature(cls.sbck.__init__).parameters.items():
                if (
                    v.default == v.empty
                    and v.kind != v.VAR_KEYWORD
                    and k != "self"
                    and k not in kwargs
                ):
                    raise ValueError(
                        f"Argument {k} is not optional for SBCK method {cls.sbck.__name__}."
                    )

            ref = ref.rename(time="time_cal")
            hist = hist.rename(time="time_cal")
            sim = sim.rename(time="time_tgt")

            if multi_dim:
                input_core_dims = [
                    ("time_cal", multi_dim),
                    ("time_cal", multi_dim),
                    ("time_tgt", multi_dim),
                ]
            else:
                input_core_dims = [("time_cal",), ("time_cal",), ("time_tgt",)]

            return xr.apply_ufunc(
                cls._apply_sbck,
                ref,
                hist,
                sim,
                input_core_dims=input_core_dims,
                kwargs={"method": cls.sbck, "fit_needs_sim": fit_needs_sim, **kwargs},
                vectorize=True,
                keep_attrs=True,
                dask="parallelized",
                output_core_dims=[input_core_dims[-1]],
                output_dtypes=[sim.dtype],
            ).rename(time_tgt="time")

        @staticmethod
        def _apply_sbck(ref, hist, sim, method, fit_needs_sim, **kwargs):
            obj = method(**kwargs)
            if fit_needs_sim:
                obj.fit(ref, hist, sim)
            else:
                obj.fit(ref, hist)
            scen = obj.predict(sim)
            if sim.ndim == 1:
                return scen[:, 0]
            return scen

    def _parse_sbck_doc(cls):
        def _parse(s):
            s = s.replace("\t", "    ")
            n = min(len(line) - len(line.lstrip()) for line in s.split("\n") if line)
            lines = []
            for line in s.split("\n"):
                line = line[n:] if line else line
                if set(line).issubset({"=", " "}):
                    line = line.replace("=", "-")
                elif set(line).issubset({"-", " "}):
                    line = line.replace("-", "~")
                lines.append(line)
            return lines

        return "\n".join(
            [
                f"SBCK_{cls.__name__}",
                "=" * (5 + len(cls.__name__)),
                (
                    f"This Adjustment object was auto-generated from the {cls.__name__} "
                    " object of package SBCK. See :ref:`Experimental wrap of SBCK`."
                ),
                "",
                (
                    "The adjust method accepts ref, hist, sim and all arguments listed "
                    'below in "Parameters". It also accepts a `multi_dim` argument '
                    "specifying the dimension across which to take the 'features' and "
                    "is valid for multivariate methods only. See :py:func:`xclim.sdba.stack_variables`."
                    "In the description below, `n_features` is the size of the `multi_dim` "
                    "dimension. There is no way of specifying parameters across other "
                    "dimensions for the moment."
                ),
                "",
                *_parse(cls.__doc__),
                *_parse(cls.__init__.__doc__),
                " Copyright(c) 2021 Yoann Robin.",
            ]
        )

    def _generate_SBCK_classes():  # noqa: N802
        classes = []
        for clsname in dir(SBCK):
            cls = getattr(SBCK, clsname)
            if (
                not clsname.startswith("_")
                and isinstance(cls, type)
                and hasattr(cls, "fit")
                and hasattr(cls, "predict")
            ):
                doc = _parse_sbck_doc(cls)
                classes.append(
                    type(
                        f"SBCK_{clsname}", (_SBCKAdjust,), {"sbck": cls, "__doc__": doc}
                    )
                )
        return classes
